# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.analysis._regrid.RectilinearRegridder`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import dask.array as da
import numpy as np
import numpy.ma as ma

from iris.analysis._regrid import RectilinearRegridder as Regridder
from iris.aux_factory import HybridHeightFactory
from iris.coord_systems import GeogCS, OSGB
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
from iris.tests.stock import global_pp, lat_lon_cube, realistic_4d


RESULT_DIR = ("analysis", "regrid")

# Convenience to access Regridder static method.
regrid = Regridder._regrid


class Test__regrid__linear(tests.IrisTest):
    def setUp(self):
        self.x = DimCoord(np.linspace(-2, 57, 60))
        self.y = DimCoord(np.linspace(0, 49, 50))
        self.xs, self.ys = np.meshgrid(self.x.points, self.y.points)

        def transformation(x, y):
            return x + y ** 2

        # Construct a function which adds dimensions to the 2D data array
        # so that we can test higher dimensional functionality.
        def dim_extender(arr):
            return arr[np.newaxis, ..., np.newaxis] * [1, 2]

        self.data = dim_extender(transformation(self.xs, self.ys))

        target_x = np.linspace(-3, 60, 4)
        target_y = np.linspace(0.5, 51, 3)
        self.target_x, self.target_y = np.meshgrid(target_x, target_y)

        #: Expected values, which not quite the analytical value, but
        #: representative of the bilinear interpolation scheme.
        self.expected = np.array(
            [
                [
                    [
                        [np.nan, np.nan],
                        [18.5, 37.0],
                        [39.5, 79.0],
                        [np.nan, np.nan],
                    ],
                    [
                        [np.nan, np.nan],
                        [681.25, 1362.5],
                        [702.25, 1404.5],
                        [np.nan, np.nan],
                    ],
                    [
                        [np.nan, np.nan],
                        [np.nan, np.nan],
                        [np.nan, np.nan],
                        [np.nan, np.nan],
                    ],
                ]
            ]
        )

        self.x_dim = 2
        self.y_dim = 1

    def assert_values(self, values):
        # values is a list of [x, y, [val1, val2]]
        xs, ys, expecteds = zip(*values)
        expecteds = np.array(expecteds)[None, None, ...]
        result = regrid(
            self.data,
            self.x_dim,
            self.y_dim,
            self.x,
            self.y,
            np.array([xs]),
            np.array([ys]),
        )
        self.assertArrayAllClose(result, expecteds, rtol=1e-04, equal_nan=True)

        # Check that transposing the input data results in the same values
        ndim = self.data.ndim
        result2 = regrid(
            self.data.T,
            ndim - self.x_dim - 1,
            ndim - self.y_dim - 1,
            self.x,
            self.y,
            np.array([xs]),
            np.array([ys]),
        )
        self.assertArrayEqual(result.T, result2)

    def test_single_values(self):
        # Check that the values are sensible e.g. (3 + 4**2 == 19)
        self.assert_values(
            [
                [3, 4, [19, 38]],
                [-2, 0, [-2, -4]],
                [-2.01, 0, [np.nan, np.nan]],
                [2, -0.01, [np.nan, np.nan]],
                [57, 0, [57, 114]],
                [57.01, 0, [np.nan, np.nan]],
                [57, 49, [2458, 4916]],
                [57, 49.01, [np.nan, np.nan]],
            ]
        )

    def test_simple_result(self):
        result = regrid(
            self.data,
            self.x_dim,
            self.y_dim,
            self.x,
            self.y,
            self.target_x,
            self.target_y,
        )
        self.assertArrayEqual(result, self.expected)

    def test_simple_masked(self):
        data = ma.MaskedArray(self.data, mask=True)
        data.mask[:, 1:30, 1:30] = False
        result = regrid(
            data,
            self.x_dim,
            self.y_dim,
            self.x,
            self.y,
            self.target_x,
            self.target_y,
        )
        expected_mask = np.array(
            [
                [
                    [[True, True], [True, True], [True, True], [True, True]],
                    [[True, True], [False, False], [True, True], [True, True]],
                    [[True, True], [True, True], [True, True], [True, True]],
                ]
            ],
            dtype=bool,
        )
        expected = ma.MaskedArray(self.expected, mask=expected_mask)
        self.assertMaskedArrayEqual(result, expected)

    def test_simple_masked_no_mask(self):
        data = ma.MaskedArray(self.data, mask=False)
        result = regrid(
            data,
            self.x_dim,
            self.y_dim,
            self.x,
            self.y,
            self.target_x,
            self.target_y,
        )
        self.assertIsInstance(result, ma.MaskedArray)

    def test_result_transpose_shape(self):
        ndim = self.data.ndim
        result = regrid(
            self.data.T,
            ndim - self.x_dim - 1,
            ndim - self.y_dim - 1,
            self.x,
            self.y,
            self.target_x,
            self.target_y,
        )
        self.assertArrayEqual(result, self.expected.T)

    def test_reverse_x_coord(self):
        index = [slice(None)] * self.data.ndim
        index[self.x_dim] = slice(None, None, -1)
        result = regrid(
            self.data[tuple(index)],
            self.x_dim,
            self.y_dim,
            self.x[::-1],
            self.y,
            self.target_x,
            self.target_y,
        )
        self.assertArrayEqual(result, self.expected)

    def test_circular_x_coord(self):
        # Check that interpolation of a circular src coordinate doesn't result
        # in an out of bounds value.
        self.x.circular = True
        self.x.units = "degree"
        result = regrid(
            self.data,
            self.x_dim,
            self.y_dim,
            self.x,
            self.y,
            np.array([[58]]),
            np.array([[0]]),
        )
        self.assertArrayAlmostEqual(
            result, np.array([56.80398671, 113.60797342], ndmin=self.data.ndim)
        )


# Check what happens to NaN values, extrapolated values, and
# masked values.
class Test__regrid__extrapolation_modes(tests.IrisTest):
    values_by_method = {
        "linear": [
            [np.nan, np.nan, 2, 3, np.nan],
            [np.nan, np.nan, 6, 7, np.nan],
            [8, 9, 10, 11, np.nan],
        ],
        "nearest": [
            [np.nan, 1, 2, 3, np.nan],
            [4, 5, 6, 7, np.nan],
            [8, 9, 10, 11, np.nan],
        ],
    }

    extrapolate_values_by_method = {
        "linear": [
            [np.nan, np.nan, 2, 3, 4],
            [np.nan, np.nan, 6, 7, 8],
            [8, 9, 10, 11, 12],
        ],
        "nearest": [[np.nan, 1, 2, 3, 3], [4, 5, 6, 7, 7], [8, 9, 10, 11, 11]],
    }

    def setUp(self):
        self.methods = ("linear", "nearest")
        self.test_dtypes = [
            np.dtype(spec)
            for spec in ("i1", "i2", "i4", "i8", "f2", "f4", "f8")
        ]

    def _regrid(self, data, method, extrapolation_mode=None):
        x = np.arange(4)
        y = np.arange(3)
        x_coord = DimCoord(x)
        y_coord = DimCoord(y)
        x_dim, y_dim = 1, 0
        grid_x, grid_y = np.meshgrid(np.arange(5), y)
        kwargs = dict(method=method)
        if extrapolation_mode is not None:
            kwargs["extrapolation_mode"] = extrapolation_mode
        result = regrid(
            data, x_dim, y_dim, x_coord, y_coord, grid_x, grid_y, **kwargs
        )
        return result

    def test_default_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> NaN
        data = np.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        for method in self.methods:
            result = self._regrid(data, method)
            self.assertNotIsInstance(result, ma.MaskedArray)
            expected = self.values_by_method[method]
            self.assertArrayEqual(result, expected)

    def test_default_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> Masked
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        data[2, 3] = ma.masked
        for method in self.methods:
            result = self._regrid(data, method)
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 1, 1]]
            values = self.values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_default_maskedarray_none_masked(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> N/A
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        for method in self.methods:
            result = self._regrid(data, method)
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 0, 1]]
            values = self.values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_default_maskedarray_none_masked_expanded(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> N/A
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        # Make sure the mask has been expanded
        data.mask = False
        data[0, 0] = np.nan
        for method in self.methods:
            result = self._regrid(data, method)
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 0, 1]]
            values = self.values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_method_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> linear
        data = np.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        for method in self.methods:
            result = self._regrid(data, method, "extrapolate")
            self.assertNotIsInstance(result, ma.MaskedArray)
            expected = self.extrapolate_values_by_method[method]
            self.assertArrayEqual(result, expected)

    def test_method_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> linear
        # Masked        -> Masked
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        data[2, 3] = ma.masked
        for method in self.methods:
            result = self._regrid(data, method, "extrapolate")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 1, 1]]
            values = self.extrapolate_values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_nan_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> NaN
        data = np.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        for method in self.methods:
            result = self._regrid(data, method, "nan")
            self.assertNotIsInstance(result, ma.MaskedArray)
            expected = self.values_by_method[method]
            self.assertArrayEqual(result, expected)

    def test_nan_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> NaN
        # Masked        -> Masked
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        data[2, 3] = ma.masked
        for method in self.methods:
            result = self._regrid(data, method, "nan")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 1, 0]]
            values = self.values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_error_ndarray(self):
        # Values irrelevant - the function raises an error.
        data = np.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        for method in self.methods:
            with self.assertRaisesRegex(ValueError, "out of bounds"):
                self._regrid(data, method, "error")

    def test_error_maskedarray(self):
        # Values irrelevant - the function raises an error.
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        data[2, 3] = ma.masked
        for method in self.methods:
            with self.assertRaisesRegex(ValueError, "out of bounds"):
                self._regrid(data, method, "error")

    def test_mask_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked (this is different from all the other
        #                          modes)
        data = np.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        for method in self.methods:
            result = self._regrid(data, method, "mask")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 0, 1]]
            values = self.values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_mask_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> Masked
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        data[2, 3] = ma.masked
        for method in self.methods:
            result = self._regrid(data, method, "mask")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 1, 1]]
            values = self.values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_nanmask_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> NaN
        data = np.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        for method in self.methods:
            result = self._regrid(data, method, "nanmask")
            self.assertNotIsInstance(result, ma.MaskedArray)
            expected = self.values_by_method[method]
            self.assertArrayEqual(result, expected)

    def test_nanmask_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> Masked
        data = ma.arange(12, dtype=np.float).reshape(3, 4)
        data[0, 0] = np.nan
        data[2, 3] = ma.masked
        for method in self.methods:
            result = self._regrid(data, method, "nanmask")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 1, 1]]
            values = self.values_by_method[method]
            expected = ma.MaskedArray(values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_invalid(self):
        data = np.arange(12, dtype=np.float).reshape(3, 4)
        emsg = "Invalid extrapolation mode"
        for method in self.methods:
            with self.assertRaisesRegex(ValueError, emsg):
                self._regrid(data, method, "BOGUS")

    def test_method_result_types(self):
        # Check return types from basic calculation on floats and ints.
        for method in self.methods:
            result_dtypes = {}
            for source_dtype in self.test_dtypes:
                data = np.arange(12, dtype=source_dtype).reshape(3, 4)
                result = self._regrid(data, method)
                result_dtypes[source_dtype] = result.dtype
            if method == "linear":
                # Linear results are promoted to float.
                expected_types_mapping = {
                    test_dtype: np.promote_types(test_dtype, np.float16)
                    for test_dtype in self.test_dtypes
                }
            if method == "nearest":
                # Nearest results are the same as the original data.
                expected_types_mapping = {
                    test_dtype: test_dtype for test_dtype in self.test_dtypes
                }
            self.assertEqual(result_dtypes, expected_types_mapping)


class Test___call___lazy(tests.IrisTest):
    def setUp(self):
        self.cube = lat_lon_cube()
        # Regridder method and extrapolation-mode.
        self.args = ("linear", "mask")
        self.regridder = Regridder(self.cube, self.cube, *self.args)
        self.lazy_cube = self.cube.copy(da.asarray(self.cube.data))
        self.lazy_regridder = Regridder(
            self.lazy_cube, self.lazy_cube, *self.args
        )

    def test_lazy_regrid(self):
        result = self.lazy_regridder(self.lazy_cube)
        self.assertTrue(result.has_lazy_data())
        expected = self.regridder(self.cube)
        self.assertTrue(result == expected)


class Test___call____invalid_types(tests.IrisTest):
    def setUp(self):
        self.cube = lat_lon_cube()
        # Regridder method and extrapolation-mode.
        self.args = ("linear", "mask")
        self.regridder = Regridder(self.cube, self.cube, *self.args)

    def test_src_as_array(self):
        arr = np.zeros((3, 4))
        with self.assertRaises(TypeError):
            Regridder(arr, self.cube, *self.args)
        with self.assertRaises(TypeError):
            self.regridder(arr)

    def test_grid_as_array(self):
        with self.assertRaises(TypeError):
            Regridder(self.cube, np.zeros((3, 4)), *self.args)

    def test_src_as_int(self):
        with self.assertRaises(TypeError):
            Regridder(42, self.cube, *self.args)
        with self.assertRaises(TypeError):
            self.regridder(42)

    def test_grid_as_int(self):
        with self.assertRaises(TypeError):
            Regridder(self.cube, 42, *self.args)


class Test___call____missing_coords(tests.IrisTest):
    def setUp(self):
        self.args = ("linear", "mask")

    def ok_bad(self, coord_names):
        # Deletes the named coords from `bad`.
        ok = lat_lon_cube()
        bad = lat_lon_cube()
        for name in coord_names:
            bad.remove_coord(name)
        return ok, bad

    def test_src_missing_lat(self):
        ok, bad = self.ok_bad(["latitude"])
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)
        regridder = Regridder(ok, ok, *self.args)
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_missing_lat(self):
        ok, bad = self.ok_bad(["latitude"])
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)

    def test_src_missing_lon(self):
        ok, bad = self.ok_bad(["longitude"])
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)
        regridder = Regridder(ok, ok, *self.args)
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_missing_lon(self):
        ok, bad = self.ok_bad(["longitude"])
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)

    def test_src_missing_lat_lon(self):
        ok, bad = self.ok_bad(["latitude", "longitude"])
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)
        regridder = Regridder(ok, ok, *self.args)
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_missing_lat_lon(self):
        ok, bad = self.ok_bad(["latitude", "longitude"])
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)


class Test___call____not_dim_coord(tests.IrisTest):
    def setUp(self):
        self.args = ("linear", "mask")

    def ok_bad(self, coord_name):
        # Demotes the named DimCoord on `bad` to an AuxCoord.
        ok = lat_lon_cube()
        bad = lat_lon_cube()
        coord = bad.coord(coord_name)
        dims = bad.coord_dims(coord)
        bad.remove_coord(coord_name)
        aux_coord = AuxCoord.from_coord(coord)
        bad.add_aux_coord(aux_coord, dims)
        return ok, bad

    def test_src_with_aux_lat(self):
        ok, bad = self.ok_bad("latitude")
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)
        regridder = Regridder(ok, ok, *self.args)
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_with_aux_lat(self):
        ok, bad = self.ok_bad("latitude")
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)

    def test_src_with_aux_lon(self):
        ok, bad = self.ok_bad("longitude")
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)
        regridder = Regridder(ok, ok, *self.args)
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_with_aux_lon(self):
        ok, bad = self.ok_bad("longitude")
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)


class Test___call____not_dim_coord_share(tests.IrisTest):
    def setUp(self):
        self.args = ("linear", "mask")

    def ok_bad(self):
        # Make lat/lon share a single dimension on `bad`.
        ok = lat_lon_cube()
        bad = lat_lon_cube()
        lat = bad.coord("latitude")
        bad = bad[0, : lat.shape[0]]
        bad.remove_coord("latitude")
        bad.add_aux_coord(lat, 0)
        return ok, bad

    def test_src_shares_dim(self):
        ok, bad = self.ok_bad()
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)
        regridder = Regridder(ok, ok, *self.args)
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_shares_dim(self):
        ok, bad = self.ok_bad()
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)


class Test___call____bad_georeference(tests.IrisTest):
    def setUp(self):
        self.args = ("linear", "mask")

    def ok_bad(self, lat_cs, lon_cs):
        # Updates `bad` to use the given coordinate systems.
        ok = lat_lon_cube()
        bad = lat_lon_cube()
        bad.coord("latitude").coord_system = lat_cs
        bad.coord("longitude").coord_system = lon_cs
        return ok, bad

    def test_src_no_cs(self):
        ok, bad = self.ok_bad(None, None)
        regridder = Regridder(bad, ok, *self.args)
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_no_cs(self):
        ok, bad = self.ok_bad(None, None)
        regridder = Regridder(ok, bad, *self.args)
        with self.assertRaises(ValueError):
            regridder(ok)

    def test_src_one_cs(self):
        ok, bad = self.ok_bad(None, GeogCS(6371000))
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)

    def test_grid_one_cs(self):
        ok, bad = self.ok_bad(None, GeogCS(6371000))
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)

    def test_src_inconsistent_cs(self):
        ok, bad = self.ok_bad(GeogCS(6370000), GeogCS(6371000))
        with self.assertRaises(ValueError):
            Regridder(bad, ok, *self.args)

    def test_grid_inconsistent_cs(self):
        ok, bad = self.ok_bad(GeogCS(6370000), GeogCS(6371000))
        with self.assertRaises(ValueError):
            Regridder(ok, bad, *self.args)


class Test___call____bad_angular_units(tests.IrisTest):
    def ok_bad(self):
        # Changes the longitude coord to radians on `bad`.
        ok = lat_lon_cube()
        bad = lat_lon_cube()
        bad.coord("longitude").units = "radians"
        return ok, bad

    def test_src_radians(self):
        ok, bad = self.ok_bad()
        regridder = Regridder(bad, ok, "linear", "mask")
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_radians(self):
        ok, bad = self.ok_bad()
        with self.assertRaises(ValueError):
            Regridder(ok, bad, "linear", "mask")


def uk_cube():
    data = np.arange(12, dtype=np.float32).reshape(3, 4)
    uk = Cube(data)
    cs = OSGB()
    y_coord = DimCoord(
        np.arange(3), "projection_y_coordinate", units="m", coord_system=cs
    )
    x_coord = DimCoord(
        np.arange(4), "projection_x_coordinate", units="m", coord_system=cs
    )
    uk.add_dim_coord(y_coord, 0)
    uk.add_dim_coord(x_coord, 1)
    surface = AuxCoord(data * 10, "surface_altitude", units="m")
    uk.add_aux_coord(surface, (0, 1))
    uk.add_aux_factory(HybridHeightFactory(orography=surface))
    return uk


class Test___call____bad_linear_units(tests.IrisTest):
    def ok_bad(self):
        # Defines `bad` with an x coordinate in km.
        ok = lat_lon_cube()
        bad = uk_cube()
        bad.coord(axis="x").units = "km"
        return ok, bad

    def test_src_km(self):
        ok, bad = self.ok_bad()
        regridder = Regridder(bad, ok, "linear", "mask")
        with self.assertRaises(ValueError):
            regridder(bad)

    def test_grid_km(self):
        ok, bad = self.ok_bad()
        with self.assertRaises(ValueError):
            Regridder(ok, bad, "linear", "mask")


class Test___call____no_coord_systems(tests.IrisTest):
    # Test behaviour in the absence of any coordinate systems.

    def setUp(self):
        self.mode = "mask"
        self.methods = ("linear", "nearest")

    def remove_coord_systems(self, cube):
        for coord in cube.coords():
            coord.coord_system = None

    def test_ok(self):
        # Ensure regridding is supported when the coordinate definitions match.
        # NB. We change the coordinate *values* to ensure that does not
        # prevent the regridding operation.
        src = uk_cube()
        self.remove_coord_systems(src)
        grid = src.copy()
        for coord in grid.dim_coords:
            coord.points = coord.points + 1
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            for coord in result.dim_coords:
                self.assertEqual(coord, grid.coord(coord))
            expected = ma.arange(12).reshape((3, 4)) + 5
            expected[:, 3] = ma.masked
            expected[2, :] = ma.masked
            self.assertMaskedArrayEqual(result.data, expected)

    def test_matching_units(self):
        # Check we are insensitive to the units provided they match.
        # NB. We change the coordinate *values* to ensure that does not
        # prevent the regridding operation.
        src = uk_cube()
        self.remove_coord_systems(src)
        # Move to unusual units (i.e. not metres or degrees).
        for coord in src.dim_coords:
            coord.units = "feet"
        grid = src.copy()
        for coord in grid.dim_coords:
            coord.points = coord.points + 1
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            for coord in result.dim_coords:
                self.assertEqual(coord, grid.coord(coord))
            expected = ma.arange(12).reshape((3, 4)) + 5
            expected[:, 3] = ma.masked
            expected[2, :] = ma.masked
            self.assertMaskedArrayEqual(result.data, expected)

    def test_different_units(self):
        src = uk_cube()
        self.remove_coord_systems(src)
        # Move to unusual units (i.e. not metres or degrees).
        for coord in src.coords():
            coord.units = "feet"
        grid = src.copy()
        grid.coord("projection_y_coordinate").units = "yards"
        # We change the coordinate *values* to ensure that does not
        # prevent the regridding operation.
        for coord in grid.dim_coords:
            coord.points = coord.points + 1
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            emsg = "matching coordinate metadata"
            with self.assertRaisesRegex(ValueError, emsg):
                regridder(src)

    def test_coord_metadata_mismatch(self):
        # Check for failure when coordinate definitions differ.
        uk = uk_cube()
        self.remove_coord_systems(uk)
        lat_lon = lat_lon_cube()
        self.remove_coord_systems(lat_lon)
        for method in self.methods:
            regridder = Regridder(uk, lat_lon, method, self.mode)
            with self.assertRaises(ValueError):
                regridder(uk)


class Test___call____extrapolation_modes(tests.IrisTest):
    values = [
        [np.nan, 6, 7, np.nan],
        [9, 10, 11, np.nan],
        [np.nan, np.nan, np.nan, np.nan],
    ]

    extrapolate_values_by_method = {
        "linear": [[np.nan, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        "nearest": [[np.nan, 6, 7, 7], [9, 10, 11, 11], [9, 10, 11, 11]],
    }

    surface_values = [
        [50, 60, 70, np.nan],
        [90, 100, 110, np.nan],
        [np.nan, np.nan, np.nan, np.nan],
    ]

    def setUp(self):
        self.methods = ("linear", "nearest")

    def _ndarray_cube(self, method):
        assert method in self.methods
        src = uk_cube()
        index = (0, 0) if method == "linear" else (1, 1)
        src.data[index] = np.nan
        return src

    def _masked_cube(self, method):
        assert method in self.methods
        src = uk_cube()
        src.data = ma.asarray(src.data)
        nan_index = (0, 0) if method == "linear" else (1, 1)
        mask_index = (2, 3)
        src.data[nan_index] = np.nan
        src.data[mask_index] = ma.masked
        return src

    def _regrid(self, src, method, extrapolation_mode="mask"):
        grid = src.copy()
        for coord in grid.dim_coords:
            coord.points = coord.points + 1
        regridder = Regridder(src, grid, method, extrapolation_mode)
        result = regridder(src)

        surface = result.coord("surface_altitude").points
        self.assertNotIsInstance(surface, ma.MaskedArray)
        self.assertArrayEqual(surface, self.surface_values)

        return result.data

    def test_default_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        for method in self.methods:
            src = self._ndarray_cube(method)
            result = self._regrid(src, method)
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 1], [0, 0, 0, 1], [1, 1, 1, 1]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_default_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> Masked
        for method in self.methods:
            src = self._masked_cube(method)
            result = self._regrid(src, method)
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 1], [0, 0, 1, 1], [1, 1, 1, 1]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_default_maskedarray_none_masked(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> N/A
        for method in self.methods:
            src = uk_cube()
            src.data = ma.asarray(src.data)
            index = (0, 0) if method == "linear" else (1, 1)
            src.data[index] = np.nan
            result = self._regrid(src, method)
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 1], [0, 0, 0, 1], [1, 1, 1, 1]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_default_maskedarray_none_masked_expanded(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> N/A
        for method in self.methods:
            src = uk_cube()
            src.data = ma.asarray(src.data)
            # Make sure the mask has been expanded
            src.data.mask = False
            index = (0, 0) if method == "linear" else (1, 1)
            src.data[index] = np.nan
            result = self._regrid(src, method)
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 1], [0, 0, 0, 1], [1, 1, 1, 1]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_method_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> linear
        for method in self.methods:
            src = self._ndarray_cube(method)
            result = self._regrid(src, method, "extrapolate")
            self.assertNotIsInstance(result, ma.MaskedArray)
            expected = self.extrapolate_values_by_method[method]
            self.assertArrayEqual(result, expected)

    def test_nan_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> NaN
        for method in self.methods:
            src = self._ndarray_cube(method)
            result = self._regrid(src, method, "nan")
            self.assertNotIsInstance(result, ma.MaskedArray)
            self.assertArrayEqual(result, self.values)

    def test_nan_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> NaN
        # Masked        -> Masked
        for method in self.methods:
            src = self._masked_cube(method)
            result = self._regrid(src, method, "nan")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_error_ndarray(self):
        # Values irrelevant - the function raises an error.
        for method in self.methods:
            src = self._ndarray_cube(method)
            with self.assertRaisesRegex(ValueError, "out of bounds"):
                self._regrid(src, method, "error")

    def test_error_maskedarray(self):
        # Values irrelevant - the function raises an error.
        for method in self.methods:
            src = self._masked_cube(method)
            with self.assertRaisesRegex(ValueError, "out of bounds"):
                self._regrid(src, method, "error")

    def test_mask_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked (this is different from all the other
        #                          modes)
        for method in self.methods:
            src = self._ndarray_cube(method)
            result = self._regrid(src, method, "mask")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 1], [0, 0, 0, 1], [1, 1, 1, 1]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_mask_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> Masked
        for method in self.methods:
            src = self._masked_cube(method)
            result = self._regrid(src, method, "mask")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 1], [0, 0, 1, 1], [1, 1, 1, 1]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_nanmask_ndarray(self):
        # NaN           -> NaN
        # Extrapolated  -> NaN
        for method in self.methods:
            src = self._ndarray_cube(method)
            result = self._regrid(src, method, "nanmask")
            self.assertNotIsInstance(result, ma.MaskedArray)
            self.assertArrayEqual(result, self.values)

    def test_nanmask_maskedarray(self):
        # NaN           -> NaN
        # Extrapolated  -> Masked
        # Masked        -> Masked
        for method in self.methods:
            src = self._masked_cube(method)
            result = self._regrid(src, method, "nanmask")
            self.assertIsInstance(result, ma.MaskedArray)
            mask = [[0, 0, 0, 1], [0, 0, 1, 1], [1, 1, 1, 1]]
            expected = ma.MaskedArray(self.values, mask)
            self.assertMaskedArrayEqual(result, expected)

    def test_invalid(self):
        src = uk_cube()
        emsg = "Invalid extrapolation mode"
        for method in self.methods:
            with self.assertRaisesRegex(ValueError, emsg):
                self._regrid(src, method, "BOGUS")


@tests.skip_data
class Test___call____rotated_to_lat_lon(tests.IrisTest):
    def setUp(self):
        self.src = realistic_4d()[:5, :2, ::40, ::30]
        self.mode = "mask"
        self.methods = ("linear", "nearest")

    def test_single_point(self):
        src = self.src[0, 0]
        grid = global_pp()[:1, :1]
        # These coordinate values have been derived by converting the
        # rotated coordinates of src[1, 1] into lat/lon by using cs2cs.
        grid.coord("longitude").points = -3.144870
        grid.coord("latitude").points = 52.406444
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            self.assertEqual(src.data[1, 1], result.data)

    def test_transposed_src(self):
        # The source dimensions are in a non-standard order.
        src = self.src
        src.transpose([3, 1, 2, 0])
        grid = self._grid_subset()
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            result.transpose([3, 1, 2, 0])
            cml = RESULT_DIR + ("{}_subset.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def _grid_subset(self):
        # The destination grid points are entirely contained within the
        # src grid points.
        grid = global_pp()[:4, :5]
        grid.coord("longitude").points = np.linspace(-3.182, -3.06, 5)
        grid.coord("latitude").points = np.linspace(52.372, 52.44, 4)
        return grid

    def test_reversed(self):
        src = self.src
        grid = self._grid_subset()

        for method in self.methods:
            cml = RESULT_DIR + ("{}_subset.cml".format(method),)
            regridder = Regridder(src, grid[::-1], method, self.mode)
            result = regridder(src)
            self.assertCMLApproxData(result[:, :, ::-1], cml)

            sample = src[:, :, ::-1]
            regridder = Regridder(sample, grid[::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, ::-1], cml)

            sample = src[:, :, :, ::-1]
            regridder = Regridder(sample, grid[::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, ::-1], cml)

            sample = src[:, :, ::-1, ::-1]
            regridder = Regridder(sample, grid[::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, ::-1], cml)

            regridder = Regridder(src, grid[:, ::-1], method, self.mode)
            result = regridder(src)
            self.assertCMLApproxData(result[:, :, :, ::-1], cml)

            sample = src[:, :, ::-1]
            regridder = Regridder(sample, grid[:, ::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, :, ::-1], cml)

            sample = src[:, :, :, ::-1]
            regridder = Regridder(sample, grid[:, ::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, :, ::-1], cml)

            sample = src[:, :, ::-1, ::-1]
            regridder = Regridder(sample, grid[:, ::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, :, ::-1], cml)

            regridder = Regridder(src, grid[::-1, ::-1], method, self.mode)
            result = regridder(src)
            self.assertCMLApproxData(result[:, :, ::-1, ::-1], cml)

            sample = src[:, :, ::-1]
            regridder = Regridder(sample, grid[::-1, ::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, ::-1, ::-1], cml)

            sample = src[:, :, :, ::-1]
            regridder = Regridder(sample, grid[::-1, ::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, ::-1, ::-1], cml)

            sample = src[:, :, ::-1, ::-1]
            regridder = Regridder(sample, grid[::-1, ::-1], method, self.mode)
            result = regridder(sample)
            self.assertCMLApproxData(result[:, :, ::-1, ::-1], cml)

    def test_grid_subset(self):
        # The destination grid points are entirely contained within the
        # src grid points.
        grid = self._grid_subset()
        for method in self.methods:
            regridder = Regridder(self.src, grid, method, self.mode)
            result = regridder(self.src)
            cml = RESULT_DIR + ("{}_subset.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def _big_grid(self):
        grid = self._grid_subset()
        big_grid = Cube(np.zeros((5, 10, 3, 4, 5)))
        big_grid.add_dim_coord(grid.coord("latitude"), 3)
        big_grid.add_dim_coord(grid.coord("longitude"), 4)
        return big_grid

    def test_grid_subset_big(self):
        # Add some extra dimensions to the destination Cube and
        # these should be safely ignored.
        big_grid = self._big_grid()
        for method in self.methods:
            regridder = Regridder(self.src, big_grid, method, self.mode)
            result = regridder(self.src)
            cml = RESULT_DIR + ("{}_subset.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def test_grid_subset_big_transposed(self):
        # The order of the grid's dimensions (including the X and Y
        # dimensions) must not affect the result.
        big_grid = self._big_grid()
        big_grid.transpose([4, 0, 3, 1, 2])
        for method in self.methods:
            regridder = Regridder(self.src, big_grid, method, self.mode)
            result = regridder(self.src)
            cml = RESULT_DIR + ("{}_subset.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def test_grid_subset_anon(self):
        # Must cope OK with anonymous source dimensions.
        src = self.src
        src.remove_coord("time")
        grid = self._grid_subset()
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            cml = RESULT_DIR + ("{}_subset_anon.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def test_grid_subset_missing_data_1(self):
        # The destination grid points are entirely contained within the
        # src grid points AND we have missing data.
        src = self.src
        src.data = ma.MaskedArray(src.data)
        src.data[:, :, 0, 0] = ma.masked
        grid = self._grid_subset()
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            cml = RESULT_DIR + ("{}_subset_masked_1.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def test_grid_subset_missing_data_2(self):
        # The destination grid points are entirely contained within the
        # src grid points AND we have missing data.
        src = self.src
        src.data = ma.MaskedArray(src.data)
        src.data[:, :, 1, 2] = ma.masked
        grid = self._grid_subset()
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            cml = RESULT_DIR + ("{}_subset_masked_2.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def test_grid_partial_overlap(self):
        # The destination grid points are partially contained within the
        # src grid points.
        grid = global_pp()[:4, :4]
        grid.coord("longitude").points = np.linspace(-3.3, -3.06, 4)
        grid.coord("latitude").points = np.linspace(52.377, 52.43, 4)
        for method in self.methods:
            regridder = Regridder(self.src, grid, method, self.mode)
            result = regridder(self.src)
            cml = RESULT_DIR + ("{}_partial_overlap.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def test_grid_no_overlap(self):
        # The destination grid points are NOT contained within the
        # src grid points.
        grid = global_pp()[:4, :4]
        grid.coord("longitude").points = np.linspace(-3.3, -3.2, 4)
        grid.coord("latitude").points = np.linspace(52.377, 52.43, 4)
        for method in self.methods:
            regridder = Regridder(self.src, grid, method, self.mode)
            result = regridder(self.src)
            self.assertCMLApproxData(result, RESULT_DIR + ("no_overlap.cml",))

    def test_grid_subset_missing_data_aux(self):
        # The destination grid points are entirely contained within the
        # src grid points AND we have missing data on the aux coordinate.
        src = self.src
        src.coord("surface_altitude").points[1, 2] = ma.masked
        grid = self._grid_subset()
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            cml = RESULT_DIR + ("{}_masked_altitude.cml".format(method),)
            self.assertCMLApproxData(result, cml)


@tests.skip_data
class Test___call____NOP(tests.IrisTest):
    def setUp(self):
        # The destination grid points are exactly the same as the
        # src grid points.
        self.src = realistic_4d()[:5, :2, ::40, ::30]
        self.lazy_src = self.src.copy(
            da.asarray(self.src.data, chunks=(1, 2) + self.src.shape[2:])
        )
        self.grid = self.src.copy()

    def test_nop__linear(self):
        regridder = Regridder(self.src, self.grid, "linear", "mask")
        result = regridder(self.src)
        self.assertEqual(result, self.src)

    def test_nop__nearest(self):
        regridder = Regridder(self.src, self.grid, "nearest", "mask")
        result = regridder(self.src)
        self.assertEqual(result, self.src)

    def test_nop__linear_lazy(self):
        regridder = Regridder(self.lazy_src, self.grid, "linear", "mask")
        result = regridder(self.lazy_src)
        self.assertEqual(result, self.lazy_src)

    def test_nop__nearest_lazy(self):
        regridder = Regridder(self.lazy_src, self.grid, "nearest", "mask")
        result = regridder(self.lazy_src)
        self.assertEqual(result, self.lazy_src)


@tests.skip_data
class Test___call____circular(tests.IrisTest):
    def setUp(self):
        src = global_pp()[::10, ::10]
        level_height = AuxCoord(
            0,
            long_name="level_height",
            units="m",
            attributes={"positive": "up"},
        )
        sigma = AuxCoord(1, long_name="sigma", units="1")
        surface_altitude = AuxCoord(
            (src.data - src.data.min()) * 50, "surface_altitude", units="m"
        )
        src.add_aux_coord(level_height)
        src.add_aux_coord(sigma)
        src.add_aux_coord(surface_altitude, [0, 1])
        hybrid_height = HybridHeightFactory(
            level_height, sigma, surface_altitude
        )
        src.add_aux_factory(hybrid_height)
        self.src = src

        grid = global_pp()[:4, :4]
        grid.coord("longitude").points = grid.coord("longitude").points - 5
        self.grid = grid
        self.mode = "mask"
        self.methods = ("linear", "nearest")

    def test_non_circular(self):
        # Non-circular src -> non-circular grid
        for method in self.methods:
            regridder = Regridder(self.src, self.grid, method, self.mode)
            result = regridder(self.src)
            self.assertFalse(result.coord("longitude").circular)
            cml = RESULT_DIR + ("{}_non_circular.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def _check_circular_results(self, src_cube, missingmask=""):
        results = []
        for method in self.methods:
            regridder = Regridder(src_cube, self.grid, method, self.mode)
            result = regridder(src_cube)
            results.append(result)
            self.assertFalse(result.coord("longitude").circular)
            cml = RESULT_DIR + (
                "{}_circular_src{}.cml".format(method, missingmask),
            )
            self.assertCMLApproxData(result, cml)
        return results

    def test_circular_src(self):
        # Circular src -> non-circular grid, standard test.
        src = self.src
        src.coord("longitude").circular = True
        self._check_circular_results(src)

    def test_circular_src__masked_missingmask(self):
        # Test the special case where src_cube.data.mask is just *False*,
        # instead of being an array.
        src = self.src
        src.coord("longitude").circular = True
        src.data = ma.MaskedArray(src.data)
        self.assertEqual(src.data.mask, False)
        method_results = self._check_circular_results(src, "missingmask")
        for method_result in method_results:
            self.assertIsInstance(method_result.data.mask, np.ndarray)
            self.assertTrue(np.all(method_result.data.mask == np.array(False)))

    def test_circular_src__masked(self):
        # Test that masked source points produce the expected masked results.

        # Define source + destination sample points.
        # Note: these are chosen to avoid any marginal edge-cases, such as
        # where a destination value matches a source point (for 'linear'), or a
        # half-way point (for 'nearest').
        src_x = [0.0, 60.0, 120.0, 180.0, 240.0, 300.0]
        dst_x = [20.0, 80.0, 140.0, 200.0, 260.0, 320.0]
        src_y = [100.0, 200.0, 300.0, 400.0, 500.0]
        dst_y = [40.0, 140.0, 240.0, 340.0, 440.0, 540.0]

        # Define the expected result masks for the tested methods,
        # when just the middle source point is masked...
        result_masks = {
            "nearest": np.array(
                [
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                ],
                dtype=bool,
            ),
            "linear": np.array(
                [
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0, 0],
                    [0, 1, 1, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                ],
                dtype=bool,
            ),
        }

        # Cook up some distinctive data values.
        src_nx, src_ny, dst_nx, dst_ny = (
            len(dd) for dd in (src_x, src_y, dst_x, dst_y)
        )
        data_x = np.arange(src_nx).reshape((1, src_nx))
        data_y = np.arange(src_ny).reshape((src_ny, 1))
        data = 3.0 + data_x + 20.0 * data_y

        # Make src and dst test cubes.
        def make_2d_cube(x_points, y_points, data):
            cube = Cube(data)
            y_coord = DimCoord(
                y_points, standard_name="latitude", units="degrees"
            )
            x_coord = DimCoord(
                x_points, standard_name="longitude", units="degrees"
            )
            x_coord.circular = True
            cube.add_dim_coord(y_coord, 0)
            cube.add_dim_coord(x_coord, 1)
            return cube

        src_cube_full = make_2d_cube(src_x, src_y, data)
        dst_cube = make_2d_cube(dst_x, dst_y, np.zeros((dst_ny, dst_nx)))

        src_cube_masked = src_cube_full.copy()
        src_cube_masked.data = ma.array(
            src_cube_masked.data, mask=np.zeros((src_ny, src_nx))
        )

        # Mask the middle source point, and give it a huge underlying data
        # value to ensure that it does not take any part in the results.
        src_cube_masked.data[2, 2] = 1e19
        src_cube_masked.data.mask[2, 2] = True

        # Test results against the unmasked operation, for each method.
        for method in self.methods:
            regridder = Regridder(
                src_cube_full, dst_cube, method, extrapolation_mode="nan"
            )
            result_basic = regridder(src_cube_full)
            result_masked = regridder(src_cube_masked)
            # Check we get a masked result
            self.assertIsInstance(result_masked.data, ma.MaskedArray)
            # Check that the result matches the basic one, except for being
            # masked at the specific expected points.
            expected_result_data = ma.array(result_basic.data)
            expected_result_data.mask = result_masks[method]
            self.assertMaskedArrayEqual(
                result_masked.data, expected_result_data
            )

    def test_circular_grid(self):
        # Non-circular src -> circular grid
        grid = self.grid
        grid.coord("longitude").circular = True
        for method in self.methods:
            regridder = Regridder(self.src, grid, method, self.mode)
            result = regridder(self.src)
            self.assertTrue(result.coord("longitude").circular)
            cml = RESULT_DIR + ("{}_circular_grid.cml".format(method),)
            self.assertCMLApproxData(result, cml)

    def test_circular_src_and_grid(self):
        # Circular src -> circular grid
        src = self.src
        src.coord("longitude").circular = True
        grid = self.grid
        grid.coord("longitude").circular = True
        for method in self.methods:
            regridder = Regridder(src, grid, method, self.mode)
            result = regridder(src)
            self.assertTrue(result.coord("longitude").circular)
            cml = RESULT_DIR + ("{}_both_circular.cml".format(method),)
            self.assertCMLApproxData(result, cml)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Miscellaneous utility functions.

"""

from collections.abc import Hashable
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
import copy
import functools
import inspect
import os
import os.path
import sys
import tempfile

import cf_units
import numpy as np
import numpy.ma as ma

import iris
from iris._deprecation import warn_deprecated
import iris.coords
import iris.exceptions
import iris.cube


def broadcast_to_shape(array, shape, dim_map):
    """
    Broadcast an array to a given shape.

    Each dimension of the array must correspond to a dimension in the
    given shape. Striding is used to repeat the array until it matches
    the desired shape, returning repeated views on the original array.
    If you need to write to the resulting array, make a copy first.

    Args:

    * array (:class:`numpy.ndarray`-like)
        An array to broadcast.

    * shape (:class:`list`, :class:`tuple` etc.):
        The shape the array should be broadcast to.

    * dim_map (:class:`list`, :class:`tuple` etc.):
        A mapping of the dimensions of *array* to their corresponding
        element in *shape*. *dim_map* must be the same length as the
        number of dimensions in *array*. Each element of *dim_map*
        corresponds to a dimension of *array* and its value provides
        the index in *shape* which the dimension of *array* corresponds
        to, so the first element of *dim_map* gives the index of *shape*
        that corresponds to the first dimension of *array* etc.

    Examples:

    Broadcasting an array of shape (2, 3) to the shape (5, 2, 6, 3)
    where the first dimension of the array corresponds to the second
    element of the desired shape and the second dimension of the array
    corresponds to the fourth element of the desired shape::

        a = np.array([[1, 2, 3], [4, 5, 6]])
        b = broadcast_to_shape(a, (5, 2, 6, 3), (1, 3))

    Broadcasting an array of shape (48, 96) to the shape (96, 48, 12)::

        # a is an array of shape (48, 96)
        result = broadcast_to_shape(a, (96, 48, 12), (1, 0))

    """
    if len(dim_map) != array.ndim:
        # We must check for this condition here because we cannot rely on
        # getting an error from numpy if the dim_map argument is not the
        # correct length, we might just get a segfault.
        raise ValueError(
            "dim_map must have an entry for every "
            "dimension of the input array"
        )

    def _broadcast_helper(a):
        strides = [0] * len(shape)
        for idim, dim in enumerate(dim_map):
            if shape[dim] != a.shape[idim]:
                # We'll get garbage values if the dimensions of array are not
                # those indicated by shape.
                raise ValueError("shape and array are not compatible")
            strides[dim] = a.strides[idim]
        return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

    array_view = _broadcast_helper(array)
    if ma.isMaskedArray(array):
        if array.mask is ma.nomask:
            # Degenerate masks can be applied as-is.
            mask_view = array.mask
        else:
            # Mask arrays need to be handled in the same way as the data array.
            mask_view = _broadcast_helper(array.mask)
        array_view = ma.array(array_view, mask=mask_view)
    return array_view


def delta(ndarray, dimension, circular=False):
    """
    Calculates the difference between values along a given dimension.

    Args:

    * ndarray:
        The array over which to do the difference.

    * dimension:
        The dimension over which to do the difference on ndarray.

    * circular:
        If not False then return n results in the requested dimension
        with the delta between the last and first element included in
        the result otherwise the result will be of length n-1 (where n
        is the length of ndarray in the given dimension's direction)

        If circular is numeric then the value of circular will be added
        to the last element of the given dimension if the last element
        is negative, otherwise the value of circular will be subtracted
        from the last element.

        The example below illustrates the process::

            original array              -180, -90,  0,    90
            delta (with circular=360):    90,  90, 90, -270+360

    .. note::

        The difference algorithm implemented is forward difference:

            >>> import numpy as np
            >>> import iris.util
            >>> original = np.array([-180, -90, 0, 90])
            >>> iris.util.delta(original, 0)
            array([90, 90, 90])
            >>> iris.util.delta(original, 0, circular=360)
            array([90, 90, 90, 90])

    """
    if circular is not False:
        _delta = np.roll(ndarray, -1, axis=dimension)
        last_element = [slice(None, None)] * ndarray.ndim
        last_element[dimension] = slice(-1, None)
        last_element = tuple(last_element)

        if not isinstance(circular, bool):
            result = np.where(ndarray[last_element] >= _delta[last_element])[0]
            _delta[last_element] -= circular
            _delta[last_element][result] += 2 * circular

        np.subtract(_delta, ndarray, _delta)
    else:
        _delta = np.diff(ndarray, axis=dimension)

    return _delta


def describe_diff(cube_a, cube_b, output_file=None):
    """
    Prints the differences that prevent compatibility between two cubes, as
    defined by :meth:`iris.cube.Cube.is_compatible()`.

    Args:

    * cube_a:
        An instance of :class:`iris.cube.Cube` or
        :class:`iris.cube.CubeMetadata`.

    * cube_b:
        An instance of :class:`iris.cube.Cube` or
        :class:`iris.cube.CubeMetadata`.

    * output_file:
        A :class:`file` or file-like object to receive output. Defaults to
        sys.stdout.

    .. seealso::

        :meth:`iris.cube.Cube.is_compatible()`

    .. note::

        Compatibility does not guarantee that two cubes can be merged.
        Instead, this function is designed to provide a verbose description
        of the differences in metadata between two cubes. Determining whether
        two cubes will merge requires additional logic that is beyond the
        scope of this function.

    """

    if output_file is None:
        output_file = sys.stdout

    if cube_a.is_compatible(cube_b):
        output_file.write("Cubes are compatible\n")
    else:
        common_keys = set(cube_a.attributes).intersection(cube_b.attributes)
        for key in common_keys:
            if np.any(cube_a.attributes[key] != cube_b.attributes[key]):
                output_file.write(
                    '"%s" cube_a attribute value "%s" is not '
                    "compatible with cube_b "
                    'attribute value "%s"\n'
                    % (key, cube_a.attributes[key], cube_b.attributes[key])
                )

        if cube_a.name() != cube_b.name():
            output_file.write(
                'cube_a name "%s" is not compatible '
                'with cube_b name "%s"\n' % (cube_a.name(), cube_b.name())
            )

        if cube_a.units != cube_b.units:
            output_file.write(
                'cube_a units "%s" are not compatible with cube_b units "%s"\n'
                % (cube_a.units, cube_b.units)
            )

        if cube_a.cell_methods != cube_b.cell_methods:
            output_file.write(
                "Cell methods\n%s\nand\n%s\nare not compatible\n"
                % (cube_a.cell_methods, cube_b.cell_methods)
            )


def guess_coord_axis(coord):
    """
    Returns a "best guess" axis name of the coordinate.

    Heuristic categorisation of the coordinate into either label
    'T', 'Z', 'Y', 'X' or None.

    Args:

    * coord:
        The :class:`iris.coords.Coord`.

    Returns:
        'T', 'Z', 'Y', 'X', or None.

    """
    axis = None

    if coord.standard_name in (
        "longitude",
        "grid_longitude",
        "projection_x_coordinate",
    ):
        axis = "X"
    elif coord.standard_name in (
        "latitude",
        "grid_latitude",
        "projection_y_coordinate",
    ):
        axis = "Y"
    elif coord.units.is_convertible("hPa") or coord.attributes.get(
        "positive"
    ) in ("up", "down"):
        axis = "Z"
    elif coord.units.is_time_reference():
        axis = "T"

    return axis


def rolling_window(a, window=1, step=1, axis=-1):
    """
    Make an ndarray with a rolling window of the last dimension

    Args:

    * a : array_like
        Array to add rolling window to

    Kwargs:

    * window : int
        Size of rolling window
    * step : int
        Size of step between rolling windows
    * axis : int
        Axis to take the rolling window over

    Returns:

        Array that is a view of the original array with an added dimension
        of the size of the given window at axis + 1.

    Examples::

        >>> x = np.arange(10).reshape((2, 5))
        >>> rolling_window(x, 3)
        array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]],
               [[5, 6, 7], [6, 7, 8], [7, 8, 9]]])

    Calculate rolling mean of last dimension::

        >>> np.mean(rolling_window(x, 3), -1)
        array([[ 1.,  2.,  3.],
               [ 6.,  7.,  8.]])

    """
    # NOTE: The implementation of this function originates from
    # https://github.com/numpy/numpy/pull/31#issuecomment-1304851 04/08/2011
    if window < 1:
        raise ValueError("`window` must be at least 1.")
    if window > a.shape[axis]:
        raise ValueError("`window` is too long.")
    if step < 1:
        raise ValueError("`step` must be at least 1.")
    axis = axis % a.ndim
    num_windows = (a.shape[axis] - window + step) // step
    shape = a.shape[:axis] + (num_windows, window) + a.shape[axis + 1 :]
    strides = (
        a.strides[:axis]
        + (step * a.strides[axis], a.strides[axis])
        + a.strides[axis + 1 :]
    )
    rw = np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)
    if ma.isMaskedArray(a):
        mask = ma.getmaskarray(a)
        strides = (
            mask.strides[:axis]
            + (step * mask.strides[axis], mask.strides[axis])
            + mask.strides[axis + 1 :]
        )
        rw = ma.array(
            rw,
            mask=np.lib.stride_tricks.as_strided(
                mask, shape=shape, strides=strides
            ),
        )
    return rw


def array_equal(array1, array2, withnans=False):
    """
    Returns whether two arrays have the same shape and elements.

    Args:

    * array1, array2 (arraylike):
        args to be compared, after normalising with :func:`np.asarray`.

    Kwargs:

    * withnans (bool):
        When unset (default), the result is False if either input contains NaN
        points.  This is the normal floating-point arithmetic result.
        When set, return True if inputs contain the same value in all elements,
        _including_ any NaN values.

    This provides much the same functionality as :func:`numpy.array_equal`, but
    with additional support for arrays of strings and NaN-tolerant operation.

    """
    array1, array2 = np.asarray(array1), np.asarray(array2)

    eq = array1.shape == array2.shape
    if eq:
        eqs = array1 == array2

        if withnans and (array1.dtype.kind == "f" or array2.dtype.kind == "f"):
            nans1, nans2 = np.isnan(array1), np.isnan(array2)
            if not np.all(nans1 == nans2):
                eq = False  # simply fail
            else:
                eqs[nans1] = True  # fix NaNs; check all the others

        if eq:
            eq = np.all(eqs)  # check equal at all points

    return eq


def approx_equal(a, b, max_absolute_error=1e-10, max_relative_error=1e-10):
    """
    Returns whether two numbers are almost equal, allowing for the
    finite precision of floating point numbers.

    """
    # Deal with numbers close to zero
    if abs(a - b) < max_absolute_error:
        return True
    # Ensure we get consistent results if "a" and "b" are supplied in the
    # opposite order.
    max_ab = max([a, b], key=abs)
    relative_error = abs(a - b) / max_ab
    return relative_error < max_relative_error


def between(lh, rh, lh_inclusive=True, rh_inclusive=True):
    """
    Provides a convenient way of defining a 3 element inequality such as
    ``a < number < b``.

    Arguments:

    * lh
        The left hand element of the inequality
    * rh
        The right hand element of the inequality

    Keywords:

    * lh_inclusive - boolean
        Affects the left hand comparison operator to use in the inequality.
        True for ``<=`` false for ``<``. Defaults to True.
    * rh_inclusive - boolean
        Same as lh_inclusive but for right hand operator.


    For example::

        between_3_and_6 = between(3, 6)
        for i in range(10):
           print(i, between_3_and_6(i))


        between_3_and_6 = between(3, 6, rh_inclusive=False)
        for i in range(10):
           print(i, between_3_and_6(i))

    """
    if lh_inclusive and rh_inclusive:
        return lambda c: lh <= c <= rh
    elif lh_inclusive and not rh_inclusive:
        return lambda c: lh <= c < rh
    elif not lh_inclusive and rh_inclusive:
        return lambda c: lh < c <= rh
    else:
        return lambda c: lh < c < rh


def reverse(cube_or_array, coords_or_dims):
    """
    Reverse the cube or array along the given dimensions.

    Args:

    * cube_or_array: :class:`iris.cube.Cube` or :class:`numpy.ndarray`
        The cube or array to reverse.
    * coords_or_dims: int, str, :class:`iris.coords.Coord` or sequence of these
        Identify one or more dimensions to reverse.  If cube_or_array is a
        numpy array, use int or a sequence of ints, as in the examples below.
        If cube_or_array is a Cube, a Coord or coordinate name (or sequence of
        these) may be specified instead.

    ::

        >>> import numpy as np
        >>> a = np.arange(24).reshape(2, 3, 4)
        >>> print(a)
        [[[ 0  1  2  3]
          [ 4  5  6  7]
          [ 8  9 10 11]]
        <BLANKLINE>
         [[12 13 14 15]
          [16 17 18 19]
          [20 21 22 23]]]
        >>> print(reverse(a, 1))
        [[[ 8  9 10 11]
          [ 4  5  6  7]
          [ 0  1  2  3]]
        <BLANKLINE>
         [[20 21 22 23]
          [16 17 18 19]
          [12 13 14 15]]]
        >>> print(reverse(a, [1, 2]))
        [[[11 10  9  8]
          [ 7  6  5  4]
          [ 3  2  1  0]]
        <BLANKLINE>
         [[23 22 21 20]
          [19 18 17 16]
          [15 14 13 12]]]

    """
    index = [slice(None, None)] * cube_or_array.ndim

    if isinstance(coords_or_dims, iris.cube.Cube):
        raise TypeError(
            "coords_or_dims must be int, str, coordinate or "
            "sequence of these.  Got cube."
        )

    if iris.cube._is_single_item(coords_or_dims):
        coords_or_dims = [coords_or_dims]

    axes = set()
    for coord_or_dim in coords_or_dims:
        if isinstance(coord_or_dim, int):
            axes.add(coord_or_dim)
        elif isinstance(cube_or_array, np.ndarray):
            raise TypeError(
                "To reverse an array, provide an int or sequence of ints."
            )
        else:
            try:
                axes.update(cube_or_array.coord_dims(coord_or_dim))
            except AttributeError:
                raise TypeError(
                    "coords_or_dims must be int, str, coordinate "
                    "or sequence of these."
                )

    axes = np.array(list(axes), ndmin=1)
    if axes.ndim != 1 or axes.size == 0:
        raise ValueError(
            "Reverse was expecting a single axis or a 1d array "
            "of axes, got %r" % axes
        )
    if np.min(axes) < 0 or np.max(axes) > cube_or_array.ndim - 1:
        raise ValueError(
            "An axis value out of range for the number of "
            "dimensions from the given array (%s) was received. "
            "Got: %r" % (cube_or_array.ndim, axes)
        )

    for axis in axes:
        index[axis] = slice(None, None, -1)

    return cube_or_array[tuple(index)]


def monotonic(array, strict=False, return_direction=False):
    """
    Return whether the given 1d array is monotonic.

    Note that, the array must not contain missing data.

    Kwargs:

    * strict (boolean)
        Flag to enable strict monotonic checking
    * return_direction (boolean)
        Flag to change return behaviour to return
        (monotonic_status, direction). Direction will be 1 for positive
        or -1 for negative. The direction is meaningless if the array is
        not monotonic.

    Returns:

    * monotonic_status (boolean)
        Whether the array was monotonic.

        If the return_direction flag was given then the returned value
        will be:

            ``(monotonic_status, direction)``

    """
    if array.ndim != 1 or len(array) <= 1:
        raise ValueError(
            "The array to check must be 1 dimensional and have "
            "more than 1 element."
        )

    if ma.isMaskedArray(array) and ma.count_masked(array) != 0:
        raise ValueError("The array to check contains missing data.")

    # Identify the directions of the largest/most-positive and
    # smallest/most-negative steps.
    d = np.diff(array)

    sign_max_d = np.sign(np.max(d))
    sign_min_d = np.sign(np.min(d))

    if strict:
        monotonic = sign_max_d == sign_min_d and sign_max_d != 0
    else:
        monotonic = (
            (sign_min_d < 0 and sign_max_d <= 0)
            or (sign_max_d > 0 and sign_min_d >= 0)
            or (sign_min_d == sign_max_d == 0)
        )

    if return_direction:
        if sign_max_d == 0:
            direction = sign_min_d
        else:
            direction = sign_max_d

        return monotonic, direction

    return monotonic


def column_slices_generator(full_slice, ndims):
    """
    Given a full slice full of tuples, return a dictionary mapping old
    data dimensions to new and a generator which gives the successive
    slices needed to index correctly (across columns).

    This routine deals with the special functionality for tuple based
    indexing e.g. [0, (3, 5), :, (1, 6, 8)] by first providing a slice
    which takes the non tuple slices out first i.e. [0, :, :, :] then
    subsequently iterates through each of the tuples taking out the
    appropriate slices i.e. [(3, 5), :, :] followed by [:, :, (1, 6, 8)]

    This method was developed as numpy does not support the direct
    approach of [(3, 5), : , (1, 6, 8)] for column based indexing.

    """
    list_of_slices = []

    # Map current dimensions to new dimensions, or None
    dimension_mapping = {None: None}
    _count_current_dim = 0
    for i, i_key in enumerate(full_slice):
        if isinstance(i_key, (int, np.integer)):
            dimension_mapping[i] = None
        else:
            dimension_mapping[i] = _count_current_dim
            _count_current_dim += 1

    # Get all of the dimensions for which a tuple of indices were provided
    # (numpy.ndarrays are treated in the same way tuples in this case)
    def is_tuple_style_index(key):
        return isinstance(key, tuple) or (
            isinstance(key, np.ndarray) and key.ndim == 1
        )

    tuple_indices = [
        i for i, key in enumerate(full_slice) if is_tuple_style_index(key)
    ]

    # stg1: Take a copy of the full_slice specification, turning all tuples
    # into a full slice
    if tuple_indices != list(range(len(full_slice))):
        first_slice = list(full_slice)
        for tuple_index in tuple_indices:
            first_slice[tuple_index] = slice(None, None)
        # turn first_slice back into a tuple ready for indexing
        first_slice = tuple(first_slice)

        list_of_slices.append(first_slice)

    # stg2 iterate over each of the tuples
    for tuple_index in tuple_indices:
        # Create a list with the indices to span the whole data array that we
        # currently have
        spanning_slice_with_tuple = [slice(None, None)] * _count_current_dim
        # Replace the slice(None, None) with our current tuple
        spanning_slice_with_tuple[dimension_mapping[tuple_index]] = full_slice[
            tuple_index
        ]

        # if we just have [(0, 1)] turn it into [(0, 1), ...] as this is
        # Numpy's syntax.
        if len(spanning_slice_with_tuple) == 1:
            spanning_slice_with_tuple.append(Ellipsis)

        spanning_slice_with_tuple = tuple(spanning_slice_with_tuple)

        list_of_slices.append(spanning_slice_with_tuple)

    # return the dimension mapping and a generator of slices
    return dimension_mapping, iter(list_of_slices)


def _build_full_slice_given_keys(keys, ndim):
    """
    Given the keys passed to a __getitem__ call, build an equivalent
    tuple of keys which span ndims.

    """
    # Ensure that we always have a tuple of keys
    if not isinstance(keys, tuple):
        keys = tuple([keys])

    # catch the case where an extra Ellipsis has been provided which can be
    # discarded iff len(keys)-1 == ndim
    if len(keys) - 1 == ndim and Ellipsis in filter(
        lambda obj: not isinstance(obj, np.ndarray), keys
    ):
        keys = list(keys)
        is_ellipsis = [key is Ellipsis for key in keys]
        keys.pop(is_ellipsis.index(True))
        keys = tuple(keys)

    # for ndim >= 1 appending a ":" to the slice specification is allowable,
    # remove this now
    if len(keys) > ndim and ndim != 0 and keys[-1] == slice(None, None):
        keys = keys[:-1]

    if len(keys) > ndim:
        raise IndexError(
            "More slices requested than dimensions. Requested "
            "%r, but there were only %s dimensions." % (keys, ndim)
        )

    # For each dimension get the slice which has been requested.
    # If no slice provided, then default to the whole dimension
    full_slice = [slice(None, None)] * ndim

    for i, key in enumerate(keys):
        if key is Ellipsis:

            # replace any subsequent Ellipsis objects in keys with
            # slice(None, None) as per Numpy
            keys = keys[:i] + tuple(
                [
                    slice(None, None) if key is Ellipsis else key
                    for key in keys[i:]
                ]
            )

            # iterate over the remaining keys in reverse to fill in
            # the gaps from the right hand side
            for j, key in enumerate(keys[:i:-1]):
                full_slice[-j - 1] = key

            # we've finished with i now so stop the iteration
            break
        else:
            full_slice[i] = key

    # remove any tuples on dimensions, turning them into numpy array's for
    # consistent behaviour
    full_slice = tuple(
        [
            np.array(key, ndmin=1) if isinstance(key, tuple) else key
            for key in full_slice
        ]
    )
    return full_slice


def _slice_data_with_keys(data, keys):
    """
    Index an array-like object as "data[keys]", with orthogonal indexing.

    Args:

    * data (array-like):
        array to index.

    * keys (list):
        list of indexes, as received from a __getitem__ call.

    This enforces an orthogonal interpretation of indexing, which means that
    both 'real' (numpy) arrays and other array-likes index in the same way,
    instead of numpy arrays doing 'fancy indexing'.

    Returns (dim_map, data_region), where :

    * dim_map (dict) :
        A dimension map, as returned by :func:`column_slices_generator`.
        i.e. "dim_map[old_dim_index]" --> "new_dim_index" or None.

    * data_region (array-like) :
        The sub-array.

    .. Note::

        Avoids copying the data, where possible.

    """
    # Combines the use of _build_full_slice_given_keys and
    # column_slices_generator.
    # By slicing on only one index at a time, this also mostly avoids copying
    # the data, except some cases when a key contains a list of indices.
    n_dims = len(data.shape)
    full_slice = _build_full_slice_given_keys(keys, n_dims)
    dims_mapping, slices_iter = column_slices_generator(full_slice, n_dims)
    for this_slice in slices_iter:
        data = data[this_slice]
        if data.ndim > 0 and min(data.shape) < 1:
            # Disallow slicings where a dimension has no points, like "[5:5]".
            raise IndexError("Cannot index with zero length slice.")

    return dims_mapping, data


def _wrap_function_for_method(function, docstring=None):
    """
    Returns a wrapper function modified to be suitable for use as a
    method.

    The wrapper function renames the first argument as "self" and allows
    an alternative docstring, thus allowing the built-in help(...)
    routine to display appropriate output.

    """
    # Generate the Python source for the wrapper function.
    # NB. The first argument is replaced with "self".
    args, varargs, varkw, defaults = inspect.getargspec(function)
    if defaults is None:
        basic_args = ["self"] + args[1:]
        default_args = []
        simple_default_args = []
    else:
        cutoff = -len(defaults)
        basic_args = ["self"] + args[1:cutoff]
        default_args = [
            "%s=%r" % pair for pair in zip(args[cutoff:], defaults)
        ]
        simple_default_args = args[cutoff:]
    var_arg = [] if varargs is None else ["*" + varargs]
    var_kw = [] if varkw is None else ["**" + varkw]
    arg_source = ", ".join(basic_args + default_args + var_arg + var_kw)
    simple_arg_source = ", ".join(
        basic_args + simple_default_args + var_arg + var_kw
    )
    source = "def %s(%s):\n    return function(%s)" % (
        function.__name__,
        arg_source,
        simple_arg_source,
    )

    # Compile the wrapper function
    # NB. There's an outstanding bug with "exec" where the locals and globals
    # dictionaries must be the same if we're to get closure behaviour.
    my_locals = {"function": function}
    exec(source, my_locals, my_locals)

    # Update the docstring if required, and return the modified function
    wrapper = my_locals[function.__name__]
    if docstring is None:
        wrapper.__doc__ = function.__doc__
    else:
        wrapper.__doc__ = docstring
    return wrapper


class _MetaOrderedHashable(ABCMeta):
    """
    A metaclass that ensures that non-abstract subclasses of _OrderedHashable
    without an explicit __init__ method are given a default __init__ method
    with the appropriate method signature.

    Also, an _init method is provided to allow subclasses with their own
    __init__ constructors to initialise their values via an explicit method
    signature.

    NB. This metaclass is used to construct the _OrderedHashable class as well
    as all its subclasses.

    """

    def __new__(cls, name, bases, namespace):
        # We only want to modify concrete classes that have defined the
        # "_names" property.
        if "_names" in namespace and not getattr(
            namespace["_names"], "__isabstractmethod__", False
        ):
            args = ", ".join(namespace["_names"])

            # Ensure the class has a constructor with explicit arguments.
            if "__init__" not in namespace:
                # Create a default __init__ method for the class
                method_source = (
                    "def __init__(self, %s):\n "
                    "self._init_from_tuple((%s,))" % (args, args)
                )
                exec(method_source, namespace)

            # Ensure the class has a "helper constructor" with explicit
            # arguments.
            if "_init" not in namespace:
                # Create a default _init method for the class
                method_source = (
                    "def _init(self, %s):\n "
                    "self._init_from_tuple((%s,))" % (args, args)
                )
                exec(method_source, namespace)

        return super().__new__(cls, name, bases, namespace)


@functools.total_ordering
class _OrderedHashable(Hashable, metaclass=_MetaOrderedHashable):
    """
    Convenience class for creating "immutable", hashable, and ordered classes.

    Instance identity is defined by the specific list of attribute names
    declared in the abstract attribute "_names". Subclasses must declare the
    attribute "_names" as an iterable containing the names of all the
    attributes relevant to equality/hash-value/ordering.

    Initial values should be set by using ::
        self._init(self, value1, value2, ..)

    .. note::

        It's the responsibility of the subclass to ensure that the values of
        its attributes are themselves hashable.

    """

    @property
    @abstractmethod
    def _names(self):
        """
        Override this attribute to declare the names of all the attributes
        relevant to the hash/comparison semantics.

        """
        pass

    def _init_from_tuple(self, values):
        for name, value in zip(self._names, values):
            object.__setattr__(self, name, value)

    def __repr__(self):
        class_name = type(self).__name__
        attributes = ", ".join(
            "%s=%r" % (name, value)
            for (name, value) in zip(self._names, self._as_tuple())
        )
        return "%s(%s)" % (class_name, attributes)

    def _as_tuple(self):
        return tuple(getattr(self, name) for name in self._names)

    # Prevent attribute updates

    def __setattr__(self, name, value):
        raise AttributeError(
            "Instances of %s are immutable" % type(self).__name__
        )

    def __delattr__(self, name):
        raise AttributeError(
            "Instances of %s are immutable" % type(self).__name__
        )

    # Provide hash semantics

    def _identity(self):
        return self._as_tuple()

    def __hash__(self):
        return hash(self._identity())

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self._identity() == other._identity()
        )

    def __ne__(self, other):
        # Since we've defined __eq__ we should also define __ne__.
        return not self == other

    # Provide default ordering semantics

    def __lt__(self, other):
        if isinstance(other, _OrderedHashable):
            return self._identity() < other._identity()
        else:
            return NotImplemented


def create_temp_filename(suffix=""):
    """Return a temporary file name.

    Args:

        * suffix  -  Optional filename extension.

    """
    temp_file = tempfile.mkstemp(suffix)
    os.close(temp_file[0])
    return temp_file[1]


def clip_string(the_str, clip_length=70, rider="..."):
    """
    Returns a clipped version of the string based on the specified clip
    length and whether or not any graceful clip points can be found.

    If the string to be clipped is shorter than the specified clip
    length, the original string is returned.

    If the string is longer than the clip length, a graceful point (a
    space character) after the clip length is searched for. If a
    graceful point is found the string is clipped at this point and the
    rider is added. If no graceful point can be found, then the string
    is clipped exactly where the user requested and the rider is added.

    Args:

    * the_str
        The string to be clipped
    * clip_length
        The length in characters that the input string should be clipped
        to. Defaults to a preconfigured value if not specified.
    * rider
        A series of characters appended at the end of the returned
        string to show it has been clipped. Defaults to a preconfigured
        value if not specified.

    Returns:
        The string clipped to the required length with a rider appended.
        If the clip length was greater than the original string, the
        original string is returned unaltered.

    """

    if clip_length >= len(the_str) or clip_length <= 0:
        return the_str
    else:
        if the_str[clip_length].isspace():
            return the_str[:clip_length] + rider
        else:
            first_part = the_str[:clip_length]
            remainder = the_str[clip_length:]

            # Try to find a graceful point at which to trim i.e. a space
            # If no graceful point can be found, then just trim where the user
            # specified by adding an empty slice of the remainder ( [:0] )
            termination_point = remainder.find(" ")
            if termination_point == -1:
                termination_point = 0

            return first_part + remainder[:termination_point] + rider


def format_array(arr):
    """
    Returns the given array as a string, using the python builtin str
    function on a piecewise basis.

    Useful for xml representation of arrays.

    For customisations, use the :mod:`numpy.core.arrayprint` directly.

    """

    summary_insert = ""
    summary_threshold = 85
    edge_items = 3
    ffunc = str
    formatArray = np.core.arrayprint._formatArray
    max_line_len = 50
    legacy = "1.13"
    if arr.size > summary_threshold:
        summary_insert = "..."
    options = np.get_printoptions()
    options["legacy"] = legacy
    with _printopts_context(**options):
        result = formatArray(
            arr,
            ffunc,
            max_line_len,
            next_line_prefix="\t\t",
            separator=", ",
            edge_items=edge_items,
            summary_insert=summary_insert,
            legacy=legacy,
        )

    return result


@contextmanager
def _printopts_context(**kwargs):
    """
    Update the numpy printoptions for the life of this context manager.

    Note: this function can be removed with numpy>=1.15 thanks to
          https://github.com/numpy/numpy/pull/10406

    """
    original_opts = np.get_printoptions()
    np.set_printoptions(**kwargs)
    try:
        yield
    finally:
        np.set_printoptions(**original_opts)


def new_axis(src_cube, scalar_coord=None):
    """
    Create a new axis as the leading dimension of the cube, promoting a scalar
    coordinate if specified.

    Args:

    * src_cube (:class:`iris.cube.Cube`)
        Source cube on which to generate a new axis.

    Kwargs:

    * scalar_coord (:class:`iris.coord.Coord` or 'string')
        Scalar coordinate to promote to a dimension coordinate.

    Returns:
        A new :class:`iris.cube.Cube` instance with one extra leading dimension
        (length 1).

    For example::

        >>> cube.shape
        (360, 360)
        >>> ncube = iris.util.new_axis(cube, 'time')
        >>> ncube.shape
        (1, 360, 360)

    """
    if scalar_coord is not None:
        scalar_coord = src_cube.coord(scalar_coord)

    # Indexing numpy arrays requires loading deferred data here returning a
    # copy of the data with a new leading dimension.
    # If the source cube is a Masked Constant, it is changed here to a Masked
    # Array to allow the mask to gain an extra dimension with the data.
    if src_cube.has_lazy_data():
        new_cube = iris.cube.Cube(src_cube.lazy_data()[None])
    else:
        if isinstance(src_cube.data, ma.core.MaskedConstant):
            new_data = ma.array([np.nan], mask=[True])
        else:
            new_data = src_cube.data[None]
        new_cube = iris.cube.Cube(new_data)

    new_cube.metadata = src_cube.metadata

    for coord in src_cube.aux_coords:
        if scalar_coord and scalar_coord == coord:
            dim_coord = iris.coords.DimCoord.from_coord(coord)
            new_cube.add_dim_coord(dim_coord, 0)
        else:
            dims = np.array(src_cube.coord_dims(coord)) + 1
            new_cube.add_aux_coord(coord.copy(), dims)

    for coord in src_cube.dim_coords:
        coord_dims = np.array(src_cube.coord_dims(coord)) + 1
        new_cube.add_dim_coord(coord.copy(), coord_dims)

    nonderived_coords = src_cube.dim_coords + src_cube.aux_coords
    coord_mapping = {
        id(old_co): new_cube.coord(old_co) for old_co in nonderived_coords
    }
    for factory in src_cube.aux_factories:
        new_factory = factory.updated(coord_mapping)
        new_cube.add_aux_factory(new_factory)

    return new_cube


def as_compatible_shape(src_cube, target_cube):
    """
    Return a cube with added length one dimensions to match the dimensionality
    and dimension ordering of `target_cube`.

    This function can be used to add the dimensions that have been collapsed,
    aggregated or sliced out, promoting scalar coordinates to length one
    dimension coordinates where necessary. It operates by matching coordinate
    metadata to infer the dimensions that need modifying, so the provided
    cubes must have coordinates with the same metadata
    (see :class:`iris.common.CoordMetadata`).

    .. note:: This function will load and copy the data payload of `src_cube`.

    .. deprecated:: 3.0.0

       Instead use :class:`~iris.common.resolve.Resolve`. For example, rather
       than calling ``as_compatible_shape(src_cube, target_cube)`` replace
       with ``Resolve(src_cube, target_cube)(target_cube.core_data())``.

    Args:

    * src_cube:
        An instance of :class:`iris.cube.Cube` with missing dimensions.

    * target_cube:
        An instance of :class:`iris.cube.Cube` with the desired dimensionality.

    Returns:
        A instance of :class:`iris.cube.Cube` with the same dimensionality as
        `target_cube` but with the data and coordinates from `src_cube`
        suitably reshaped to fit.

    """
    wmsg = (
        "iris.util.as_compatible_shape has been deprecated and will be "
        "removed, please use iris.common.resolve.Resolve instead."
    )
    warn_deprecated(wmsg)

    dim_mapping = {}
    for coord in target_cube.aux_coords + target_cube.dim_coords:
        dims = target_cube.coord_dims(coord)
        try:
            collapsed_dims = src_cube.coord_dims(coord)
        except iris.exceptions.CoordinateNotFoundError:
            continue
        if collapsed_dims:
            if len(collapsed_dims) == len(dims):
                for dim_from, dim_to in zip(dims, collapsed_dims):
                    dim_mapping[dim_from] = dim_to
        elif dims:
            for dim_from in dims:
                dim_mapping[dim_from] = None

    if len(dim_mapping) != target_cube.ndim:
        raise ValueError(
            "Insufficient or conflicting coordinate "
            "metadata. Cannot infer dimension mapping "
            "to restore cube dimensions."
        )

    new_shape = [1] * target_cube.ndim
    for dim_from, dim_to in dim_mapping.items():
        if dim_to is not None:
            new_shape[dim_from] = src_cube.shape[dim_to]

    new_data = src_cube.data.copy()

    # Transpose the data (if necessary) to prevent assignment of
    # new_shape doing anything except adding length one dims.
    order = [v for k, v in sorted(dim_mapping.items()) if v is not None]
    if order != sorted(order):
        new_order = [order.index(i) for i in range(len(order))]
        new_data = np.transpose(new_data, new_order).copy()

    new_cube = iris.cube.Cube(new_data.reshape(new_shape))
    new_cube.metadata = copy.deepcopy(src_cube.metadata)

    # Record a mapping from old coordinate IDs to new coordinates,
    # for subsequent use in creating updated aux_factories.
    coord_mapping = {}

    reverse_mapping = {v: k for k, v in dim_mapping.items() if v is not None}

    def add_coord(coord):
        """Closure used to add a suitably reshaped coord to new_cube."""
        all_dims = target_cube.coord_dims(coord)
        src_dims = [
            dim
            for dim in src_cube.coord_dims(coord)
            if src_cube.shape[dim] > 1
        ]
        mapped_dims = [reverse_mapping[dim] for dim in src_dims]
        length1_dims = [dim for dim in all_dims if new_cube.shape[dim] == 1]
        dims = length1_dims + mapped_dims
        shape = [new_cube.shape[dim] for dim in dims]
        if not shape:
            shape = [1]
        points = coord.points.reshape(shape)
        bounds = None
        if coord.has_bounds():
            bounds = coord.bounds.reshape(shape + [coord.nbounds])
        new_coord = coord.copy(points=points, bounds=bounds)
        # If originally in dim_coords, add to dim_coords, otherwise add to
        # aux_coords.
        if target_cube.coords(coord, dim_coords=True):
            try:
                new_cube.add_dim_coord(new_coord, dims)
            except ValueError:
                # Catch cases where the coord is an AuxCoord and therefore
                # cannot be added to dim_coords.
                new_cube.add_aux_coord(new_coord, dims)
        else:
            new_cube.add_aux_coord(new_coord, dims)
        coord_mapping[id(coord)] = new_coord

    for coord in src_cube.aux_coords + src_cube.dim_coords:
        add_coord(coord)
    for factory in src_cube.aux_factories:
        new_cube.add_aux_factory(factory.updated(coord_mapping))

    return new_cube


def squeeze(cube):
    """
    Removes any dimension of length one. If it has an associated DimCoord or
    AuxCoord, this becomes a scalar coord.

    Args:

    * cube (:class:`iris.cube.Cube`)
        Source cube to remove length 1 dimension(s) from.

    Returns:
        A new :class:`iris.cube.Cube` instance without any dimensions of
        length 1.

    For example::

        >>> cube.shape
        (1, 360, 360)
        >>> ncube = iris.util.squeeze(cube)
        >>> ncube.shape
        (360, 360)

    """

    slices = [
        0 if cube.shape[dim] == 1 else slice(None) for dim in range(cube.ndim)
    ]

    squeezed = cube[tuple(slices)]

    return squeezed


def file_is_newer_than(result_path, source_paths):
    """
    Return whether the 'result' file has a later modification time than all of
    the 'source' files.

    If a stored result depends entirely on known 'sources', it need only be
    re-built when one of them changes.  This function can be used to test that
    by comparing file timestamps.

    Args:

    * result_path (string):
        The filepath of a file containing some derived result data.
    * source_paths (string or iterable of strings):
        The path(s) to the original datafiles used to make the result.  May
        include wildcards and '~' expansions (like Iris load paths), but not
        URIs.

    Returns:
        True if all the sources are older than the result, else False.

        If any of the file paths describes no existing files, an exception will
        be raised.

    .. note::
        There are obvious caveats to using file timestamps for this, as correct
        usage depends on how the sources might change.  For example, a file
        could be replaced by one of the same name, but an older timestamp.

        If wildcards and '~' expansions are used, this introduces even more
        uncertainty, as then you cannot even be sure that the resulting list of
        file names is the same as the originals.  For example, some files may
        have been deleted or others added.

    .. note::
        The result file may often be a :mod:`pickle` file.  In that case, it
        also depends on the relevant module sources, so extra caution is
        required.  Ideally, an additional check on iris.__version__ is advised.

    """
    # Accept a string as a single source path
    if isinstance(source_paths, str):
        source_paths = [source_paths]
    # Fix our chosen timestamp function
    file_date = os.path.getmtime
    # Get the 'result file' time
    result_timestamp = file_date(result_path)
    # Get all source filepaths, with normal Iris.io load helper function
    source_file_paths = iris.io.expand_filespecs(source_paths)
    # Compare each filetime, for each spec, with the 'result time'
    for path in source_file_paths:
        source_timestamp = file_date(path)
        if source_timestamp >= result_timestamp:
            return False
    return True


def is_regular(coord):
    """Determine if the given coord is regular."""
    try:
        regular_step(coord)
    except iris.exceptions.CoordinateNotRegularError:
        return False
    except (TypeError, ValueError):
        return False
    return True


def regular_step(coord):
    """Return the regular step from a coord or fail."""
    if coord.ndim != 1:
        raise iris.exceptions.CoordinateMultiDimError("Expected 1D coord")
    if coord.shape[0] < 2:
        raise ValueError("Expected a non-scalar coord")

    avdiff, regular = points_step(coord.points)
    if not regular:
        msg = "Coord %s is not regular" % coord.name()
        raise iris.exceptions.CoordinateNotRegularError(msg)
    return avdiff.astype(coord.points.dtype)


def points_step(points):
    """Determine whether a NumPy array has a regular step."""
    diffs = np.diff(points)
    avdiff = np.mean(diffs)
    # TODO: This value for `rtol` is set for test_analysis to pass...
    regular = np.allclose(diffs, avdiff, rtol=0.001)
    return avdiff, regular


def unify_time_units(cubes):
    """
    Performs an in-place conversion of the time units of all time coords in the
    cubes in a given iterable. One common epoch is defined for each calendar
    found in the cubes to prevent units being defined with inconsistencies
    between epoch and calendar.

    Each epoch is defined from the first suitable time coordinate found in the
    input cubes.

    Arg:

    * cubes:
        An iterable containing :class:`iris.cube.Cube` instances.

    """
    epochs = {}

    for cube in cubes:
        for time_coord in cube.coords():
            if time_coord.units.is_time_reference():
                epoch = epochs.setdefault(
                    time_coord.units.calendar, time_coord.units.origin
                )
                new_unit = cf_units.Unit(epoch, time_coord.units.calendar)
                time_coord.convert_units(new_unit)


def _is_circular(points, modulus, bounds=None):
    """
    Determine whether the provided points or bounds are circular in nature
    relative to the modulus value.

    If the bounds are provided then these are checked for circularity rather
    than the points.

    Args:

    * points:
        :class:`numpy.ndarray` of point values.

    * modulus:
        Circularity modulus value.

    Kwargs:

    * bounds:
        :class:`numpy.ndarray` of bound values.

    Returns:
        Boolean.

    """
    circular = False
    if bounds is not None:
        # Set circular to True if the bounds ends are equivalent.
        first_bound = last_bound = None
        if bounds.ndim == 1 and bounds.shape[-1] == 2:
            first_bound = bounds[0] % modulus
            last_bound = bounds[1] % modulus
        elif bounds.ndim == 2 and bounds.shape[-1] == 2:
            first_bound = bounds[0, 0] % modulus
            last_bound = bounds[-1, 1] % modulus

        if first_bound is not None and last_bound is not None:
            circular = np.allclose(first_bound, last_bound, rtol=1.0e-5)
    else:
        # set circular if points are regular and last+1 ~= first
        if len(points) > 1:
            diffs = list(set(np.diff(points)))
            diff = np.mean(diffs)
            abs_tol = np.abs(diff * 1.0e-4)
            diff_approx_equal = np.max(np.abs(diffs - diff)) < abs_tol
            if diff_approx_equal:
                circular_value = (points[-1] + diff) % modulus
                try:
                    np.testing.assert_approx_equal(
                        points[0], circular_value, significant=4
                    )
                    circular = True
                except AssertionError:
                    if points[0] == 0:
                        try:
                            np.testing.assert_approx_equal(
                                modulus, circular_value, significant=4
                            )
                            circular = True
                        except AssertionError:
                            pass
        else:
            # XXX - Inherited behaviour from NetCDF PyKE rules.
            # We need to decide whether this is valid!
            circular = points[0] >= modulus
    return circular


def promote_aux_coord_to_dim_coord(cube, name_or_coord):
    """
    Promotes an AuxCoord on the cube to a DimCoord. This AuxCoord must be
    associated with a single cube dimension. If the AuxCoord is associated
    with a dimension that already has a DimCoord, that DimCoord gets
    demoted to an AuxCoord.

    Args:

    * cube
        An instance of :class:`iris.cube.Cube`

    * name_or_coord:
        Either

        (a) An instance of :class:`iris.coords.AuxCoord`

        or

        (b) the :attr:`standard_name`, :attr:`long_name`, or
        :attr:`var_name` of an instance of an instance of
        :class:`iris.coords.AuxCoord`.

    For example::

        >>> print cube
        air_temperature / (K)       (time: 12; latitude: 73; longitude: 96)
             Dimension coordinates:
                  time                    x      -              -
                  latitude                -      x              -
                  longitude               -      -              x
             Auxiliary coordinates:
                  year                    x      -              -
        >>> promote_aux_coord_to_dim_coord(cube, 'year')
        >>> print cube
        air_temperature / (K)       (year: 12; latitude: 73; longitude: 96)
             Dimension coordinates:
                  year                    x      -              -
                  latitude                -      x              -
                  longitude               -      -              x
             Auxiliary coordinates:
                  time                    x      -              -

    """

    if isinstance(name_or_coord, str):
        aux_coord = cube.coord(name_or_coord)
    elif isinstance(name_or_coord, iris.coords.Coord):
        aux_coord = name_or_coord
    else:
        # Don't know how to handle this type
        msg = (
            "Don't know how to handle coordinate of type {}. "
            "Ensure all coordinates are of type str or "
            "iris.coords.Coord."
        )
        msg = msg.format(type(name_or_coord))
        raise TypeError(msg)

    if aux_coord in cube.dim_coords:
        # nothing to do
        return

    if aux_coord not in cube.aux_coords:
        msg = (
            "Attempting to promote an AuxCoord ({}) "
            "which does not exist in the cube."
        )
        msg = msg.format(aux_coord.name())
        raise ValueError(msg)

    coord_dim = cube.coord_dims(aux_coord)

    if len(coord_dim) != 1:
        msg = (
            "Attempting to promote an AuxCoord ({}) "
            "which is associated with {} dimensions."
        )
        msg = msg.format(aux_coord.name(), len(coord_dim))
        raise ValueError(msg)

    try:
        dim_coord = iris.coords.DimCoord.from_coord(aux_coord)
    except ValueError as valerr:
        msg = (
            "Attempt to promote an AuxCoord ({}) fails "
            "when attempting to create a DimCoord from the "
            "AuxCoord because: {}"
        )
        msg = msg.format(aux_coord.name(), str(valerr))
        raise ValueError(msg)

    old_dim_coord = cube.coords(
        dim_coords=True, contains_dimension=coord_dim[0]
    )

    if len(old_dim_coord) == 1:
        demote_dim_coord_to_aux_coord(cube, old_dim_coord[0])

    # order matters here: don't want to remove
    # the aux_coord before have tried to make
    # dim_coord in case that fails
    cube.remove_coord(aux_coord)

    cube.add_dim_coord(dim_coord, coord_dim)


def demote_dim_coord_to_aux_coord(cube, name_or_coord):
    """
    Demotes a dimension coordinate  on the cube to an auxiliary coordinate.

    The DimCoord is demoted to an auxiliary coordinate on the cube.
    The dimension of the cube that was associated with the DimCoord becomes
    anonymous.  The class of the coordinate is left as DimCoord, it is not
    recast as an AuxCoord instance.

    Args:

    * cube
        An instance of :class:`iris.cube.Cube`

    * name_or_coord:
        Either

        (a) An instance of :class:`iris.coords.DimCoord`

        or

        (b) the :attr:`standard_name`, :attr:`long_name`, or
        :attr:`var_name` of an instance of an instance of
        :class:`iris.coords.DimCoord`.

    For example::

        >>> print cube
        air_temperature / (K)       (time: 12; latitude: 73; longitude: 96)
             Dimension coordinates:
                  time                    x      -              -
                  latitude                -      x              -
                  longitude               -      -              x
             Auxiliary coordinates:
                  year                    x      -              -
        >>> demote_dim_coord_to_aux_coord(cube, 'time')
        >>> print cube
        air_temperature / (K)        (-- : 12; latitude: 73; longitude: 96)
             Dimension coordinates:
                  latitude                -      x              -
                  longitude               -      -              x
             Auxiliary coordinates:
                  time                    x      -              -
                  year                    x      -              -

    """

    if isinstance(name_or_coord, str):
        dim_coord = cube.coord(name_or_coord)
    elif isinstance(name_or_coord, iris.coords.Coord):
        dim_coord = name_or_coord
    else:
        # Don't know how to handle this type
        msg = (
            "Don't know how to handle coordinate of type {}. "
            "Ensure all coordinates are of type str or "
            "iris.coords.Coord."
        )
        msg = msg.format(type(name_or_coord))
        raise TypeError(msg)

    if dim_coord not in cube.dim_coords:
        # nothing to do
        return

    coord_dim = cube.coord_dims(dim_coord)

    cube.remove_coord(dim_coord)

    cube.add_aux_coord(dim_coord, coord_dim)


@functools.wraps(np.meshgrid)
def _meshgrid(*xi, **kwargs):
    """
    @numpy v1.13, the dtype of each output nD coordinate is the same as its
    associated input 1D coordinate. This is not the case prior to numpy v1.13,
    where the output dtype is cast up to its highest resolution, regardlessly.

    This convenience function ensures consistent meshgrid behaviour across
    numpy versions.

    Reference: https://github.com/numpy/numpy/pull/5302

    """
    mxi = np.meshgrid(*xi, **kwargs)
    for i, (mxii, xii) in enumerate(zip(mxi, xi)):
        if mxii.dtype != xii.dtype:
            mxi[i] = mxii.astype(xii.dtype)
    return mxi


def find_discontiguities(cube, rel_tol=1e-5, abs_tol=1e-8):
    """
    Searches coord for discontiguities in the bounds array, returned as a
    boolean array (True where discontiguities are present).

    Args:

    * cube (`iris.cube.Cube`):
        The cube to be checked for discontinuities in its 'x' and 'y'
        coordinates.

    Kwargs:

    * rel_tol (float):
        The relative equality tolerance to apply in coordinate bounds
        checking.

    * abs_tol (float):
        The absolute value tolerance to apply in coordinate bounds
        checking.

    Returns:

    * result (`numpy.ndarray` of bool) :
        true/false map of which cells in the cube XY grid have
        discontiguities in the coordinate points array.

        This can be used as the input array for
        :func:`iris.util.mask_cube`.

    Examples::

        # Find any unknown discontiguities in your cube's x and y arrays:
        discontiguities = iris.util.find_discontiguities(cube)

        # Pass the resultant boolean array to `iris.util.mask_cube`
        # with a cube slice; this will use the boolean array to mask
        # any discontiguous data points before plotting:
        masked_cube_slice = iris.util.mask_cube(cube[0], discontiguities)

        # Plot the masked cube slice:
        iplt.pcolormesh(masked_cube_slice)

    """
    lats_and_lons = [
        "latitude",
        "grid_latitude",
        "longitude",
        "grid_longitude",
    ]
    spatial_coords = [
        coord for coord in cube.aux_coords if coord.name() in lats_and_lons
    ]
    dim_err_msg = (
        "Discontiguity searches are currently only supported for "
        "2-dimensional coordinates."
    )
    if len(spatial_coords) != 2:
        raise NotImplementedError(dim_err_msg)

    # Check which dimensions are spanned by each coordinate.
    for coord in spatial_coords:
        if coord.ndim != 2:
            raise NotImplementedError(dim_err_msg)
        else:
            span = set(cube.coord_dims(coord))
        if not span:
            msg = "The coordinate {!r} doesn't span a data dimension."
            raise ValueError(msg.format(coord.name()))

    # Check that the 2d coordinate arrays are the same shape as each other
    if len(spatial_coords) == 2:
        assert spatial_coords[0].points.shape == spatial_coords[1].points.shape

    # Set up unmasked boolean array the same size as the coord points array:
    bad_points_boolean = np.zeros(spatial_coords[0].points.shape, dtype=bool)

    for coord in spatial_coords:
        _, (diffs_x, diffs_y) = coord._discontiguity_in_bounds(
            rtol=rel_tol, atol=abs_tol
        )

        bad_points_boolean[:, :-1] = np.logical_or(
            bad_points_boolean[:, :-1], diffs_x
        )
        # apply mask for y-direction discontiguities:
        bad_points_boolean[:-1, :] = np.logical_or(
            bad_points_boolean[:-1, :], diffs_y
        )
    return bad_points_boolean


def mask_cube(cube, points_to_mask):
    """
    Masks any cells in the data array which correspond to cells marked `True`
    in the `points_to_mask` array.

    Args:

    * cube (`iris.cube.Cube`):
        A 2-dimensional instance of :class:`iris.cube.Cube`.

    * points_to_mask (`numpy.ndarray` of bool):
        A 2d boolean array of Truth values representing points to mask in the
        x and y arrays of the cube.

    Returns:

    * result (`iris.cube.Cube`):
        A cube whose data array is masked at points specified by input array.

    """
    cube.data = ma.masked_array(cube.data)
    cube.data[points_to_mask] = ma.masked
    return cube


def equalise_attributes(cubes):
    """
    Delete cube attributes that are not identical over all cubes in a group.

    This function simply deletes any attributes which are not the same for
    all the given cubes.  The cubes will then have identical attributes.  The
    given cubes are modified in-place.

    Args:

    * cubes (iterable of :class:`iris.cube.Cube`):
        A collection of cubes to compare and adjust.

    """
    # Work out which attributes are identical across all the cubes.
    common_keys = list(cubes[0].attributes.keys())
    for cube in cubes[1:]:
        cube_keys = list(cube.attributes.keys())
        common_keys = [
            key
            for key in common_keys
            if (
                key in cube_keys
                and np.all(cube.attributes[key] == cubes[0].attributes[key])
            )
        ]

    # Remove all the other attributes.
    for cube in cubes:
        for key in list(cube.attributes.keys()):
            if key not in common_keys:
                del cube.attributes[key]


def _strip_metadata_from_dims(cube, dims):
    """
    Remove ancillary variables and cell measures that map to specific dimensions.

    Returns a cube copy with (possibly) some cell-measures and ancillary variables removed.

    To be used by operations that modify or remove dimensions.
    Note: does nothing to (aux)-coordinates.  Those would be handled explicitly by the calling operation.

    """
    reduced_cube = cube.copy()

    # Remove any ancillary variables that span the dimension(s) being collapsed
    for ancil in reduced_cube.ancillary_variables():
        ancil_dims = reduced_cube.ancillary_variable_dims(ancil)
        if set(dims).intersection(ancil_dims):
            reduced_cube.remove_ancillary_variable(ancil)

    # Remove any cell measures that span the dimension(s) being collapsed
    for cm in reduced_cube.cell_measures():
        cm_dims = reduced_cube.cell_measure_dims(cm)
        if set(dims).intersection(cm_dims):
            reduced_cube.remove_cell_measure(cm)

    return reduced_cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :func:`iris.common.metadata._hexdigest`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy.ma as ma
import numpy as np
from xxhash import xxh64, xxh64_hexdigest

from iris.common.metadata import _hexdigest as hexdigest


class TestBytesLikeObject(tests.IrisTest):
    def setUp(self):
        self.hasher = xxh64()
        self.hasher.reset()

    @staticmethod
    def _ndarray(value):
        parts = str((value.shape, xxh64_hexdigest(value)))
        return xxh64_hexdigest(parts)

    @staticmethod
    def _masked(value):
        parts = str(
            (
                value.shape,
                xxh64_hexdigest(value.data),
                xxh64_hexdigest(value.mask),
            )
        )
        return xxh64_hexdigest(parts)

    def test_string(self):
        value = "hello world"
        self.hasher.update(value)
        expected = self.hasher.hexdigest()
        self.assertEqual(expected, hexdigest(value))

    def test_numpy_array_int(self):
        value = np.arange(10, dtype=np.int)
        expected = self._ndarray(value)
        self.assertEqual(expected, hexdigest(value))

    def test_numpy_array_float(self):
        value = np.arange(10, dtype=np.float)
        expected = self._ndarray(value)
        self.assertEqual(expected, hexdigest(value))

    def test_numpy_array_float_not_int(self):
        ivalue = np.arange(10, dtype=np.int)
        fvalue = np.arange(10, dtype=np.float)
        expected = self._ndarray(ivalue)
        self.assertNotEqual(expected, hexdigest(fvalue))

    def test_numpy_array_reshape(self):
        value = np.arange(10).reshape(2, 5)
        expected = self._ndarray(value)
        self.assertEqual(expected, hexdigest(value))

    def test_numpy_array_reshape_not_flat(self):
        value = np.arange(10).reshape(2, 5)
        expected = self._ndarray(value)
        self.assertNotEqual(expected, hexdigest(value.flatten()))

    def test_masked_array_int(self):
        value = ma.arange(10, dtype=np.int)
        expected = self._masked(value)
        self.assertEqual(expected, hexdigest(value))

        value[0] = ma.masked
        self.assertNotEqual(expected, hexdigest(value))
        expected = self._masked(value)
        self.assertEqual(expected, hexdigest(value))

    def test_masked_array_float(self):
        value = ma.arange(10, dtype=np.float)
        expected = self._masked(value)
        self.assertEqual(expected, hexdigest(value))

        value[0] = ma.masked
        self.assertNotEqual(expected, hexdigest(value))
        expected = self._masked(value)
        self.assertEqual(expected, hexdigest(value))

    def test_masked_array_float_not_int(self):
        ivalue = ma.arange(10, dtype=np.int)
        fvalue = ma.arange(10, dtype=np.float)
        expected = self._masked(ivalue)
        self.assertNotEqual(expected, hexdigest(fvalue))

    def test_masked_array_not_array(self):
        value = ma.arange(10)
        expected = self._masked(value)
        self.assertNotEqual(expected, hexdigest(value.data))

    def test_masked_array_reshape(self):
        value = ma.arange(10).reshape(2, 5)
        expected = self._masked(value)
        self.assertEqual(expected, hexdigest(value))

    def test_masked_array_reshape_not_flat(self):
        value = ma.arange(10).reshape(2, 5)
        expected = self._masked(value)
        self.assertNotEqual(expected, hexdigest(value.flatten()))


class TestNotBytesLikeObject(tests.IrisTest):
    def _expected(self, value):
        parts = str((type(value), value))
        return xxh64_hexdigest(parts)

    def test_int(self):
        value = 123
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_numpy_int(self):
        value = np.int(123)
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_float(self):
        value = 123.4
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_numpy_float(self):
        value = np.float(123.4)
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_list(self):
        value = [1, 2, 3]
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_tuple(self):
        value = (1, 2, 3)
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_dict(self):
        value = dict(one=1, two=2, three=3)
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_sentinel(self):
        value = mock.sentinel.value
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_instance(self):
        class Dummy:
            pass

        value = Dummy()
        expected = self._expected(value)
        self.assertEqual(expected, hexdigest(value))

    def test_int_not_str(self):
        value = 123
        expected = self._expected(value)
        self.assertNotEqual(expected, hexdigest(str(value)))


if __name__ == "__main__":
    tests.main()



import iris
import iris.analysis
import iris.plot as iplt
import matplotlib.pyplot as plt

global_air_temp = iris.load_cube(iris.sample_data_path("air_temp.pp"))
rotated_psl = iris.load_cube(iris.sample_data_path("rotated_pole.nc"))

rotated_air_temp = global_air_temp.regrid(rotated_psl, iris.analysis.Linear())


plt.figure(figsize=(4, 3))

iplt.pcolormesh(rotated_air_temp, norm=plt.Normalize(260, 300))
plt.title("Air temperature\n" "on a limited area rotated pole grid")
ax = plt.gca()
ax.coastlines(resolution="50m")
ax.gridlines()

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
A package for provisioning common Iris infrastructure.

"""

from .lenient import *
from .metadata import *
from .mixin import *
from .resolve import *



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris.fileformats.pp._data_bytes_to_shaped_array` function.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import io
from unittest import mock

import numpy as np
import numpy.ma as ma

import iris.fileformats.pp as pp


class Test__data_bytes_to_shaped_array__lateral_boundary_compression(
    tests.IrisTest
):
    def setUp(self):
        self.data_shape = 30, 40
        y_halo, x_halo, rim = 2, 3, 4

        data_len = np.prod(self.data_shape)
        decompressed = np.arange(data_len).reshape(*self.data_shape)
        decompressed *= np.arange(self.data_shape[1]) % 3 + 1

        decompressed_mask = np.zeros(self.data_shape, np.bool)
        decompressed_mask[
            y_halo + rim : -(y_halo + rim), x_halo + rim : -(x_halo + rim)
        ] = True

        self.decompressed = ma.masked_array(
            decompressed, mask=decompressed_mask
        )

        self.north = decompressed[-(y_halo + rim) :, :]
        self.east = decompressed[
            y_halo + rim : -(y_halo + rim), -(x_halo + rim) :
        ]
        self.south = decompressed[: y_halo + rim, :]
        self.west = decompressed[
            y_halo + rim : -(y_halo + rim), : x_halo + rim
        ]

        # Get the bytes of the north, east, south, west arrays combined.
        buf = io.BytesIO()
        buf.write(self.north.copy())
        buf.write(self.east.copy())
        buf.write(self.south.copy())
        buf.write(self.west.copy())
        buf.seek(0)
        self.data_payload_bytes = buf.read()

    def test_boundary_decompression(self):
        boundary_packing = mock.Mock(rim_width=4, x_halo=3, y_halo=2)
        lbpack = mock.Mock(n1=0)
        r = pp._data_bytes_to_shaped_array(
            self.data_payload_bytes,
            lbpack,
            boundary_packing,
            self.data_shape,
            self.decompressed.dtype,
            -9223372036854775808,
        )
        r = ma.masked_array(r, np.isnan(r), fill_value=-9223372036854775808)
        self.assertMaskedArrayEqual(r, self.decompressed)


class Test__data_bytes_to_shaped_array__land_packed(tests.IrisTest):
    def setUp(self):
        # Sets up some useful arrays for use with the land/sea mask
        # decompression.
        self.land = np.array(
            [[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1]], dtype=np.float64
        )
        sea = ~self.land.astype(np.bool)
        self.land_masked_data = np.array([1, 3, 4.5])
        self.sea_masked_data = np.array([1, 3, 4.5, -4, 5, 0, 1, 2, 3])

        # Compute the decompressed land mask data.
        self.decomp_land_data = ma.masked_array(
            [[0, 1, 0, 0], [3, 0, 0, 0], [0, 0, 0, 4.5]],
            mask=sea,
            dtype=np.float64,
        )
        # Compute the decompressed sea mask data.
        self.decomp_sea_data = ma.masked_array(
            [[1, -10, 3, 4.5], [-10, -4, 5, 0], [1, 2, 3, -10]],
            mask=self.land,
            dtype=np.float64,
        )

        self.land_mask = mock.Mock(
            data=self.land, lbrow=self.land.shape[0], lbnpt=self.land.shape[1]
        )

    def create_lbpack(self, value):
        name_mapping = dict(n5=slice(4, None), n4=3, n3=2, n2=1, n1=0)
        return pp.SplittableInt(value, name_mapping)

    def test_no_land_mask(self):
        # Check that without a mask, it returns the raw (compressed) data.
        with mock.patch("numpy.frombuffer", return_value=np.arange(3)):
            result = pp._data_bytes_to_shaped_array(
                mock.Mock(),
                self.create_lbpack(120),
                None,
                (3, 4),
                np.dtype(">f4"),
                -999,
                mask=None,
            )
            self.assertArrayAllClose(result, np.arange(3))

    def test_land_mask(self):
        # Check basic land unpacking.
        field_data = self.land_masked_data
        result = self.check_read_data(field_data, 120, self.land_mask)
        self.assertMaskedArrayEqual(result, self.decomp_land_data)

    def test_land_masked_data_too_long(self):
        # Check land unpacking with field data that is larger than the mask.
        field_data = np.tile(self.land_masked_data, 2)
        result = self.check_read_data(field_data, 120, self.land_mask)
        self.assertMaskedArrayEqual(result, self.decomp_land_data)

    def test_sea_mask(self):
        # Check basic land unpacking.
        field_data = self.sea_masked_data
        result = self.check_read_data(field_data, 220, self.land_mask)
        self.assertMaskedArrayEqual(result, self.decomp_sea_data)

    def test_sea_masked_data_too_long(self):
        # Check sea unpacking with field data that is larger than the mask.
        field_data = np.tile(self.sea_masked_data, 2)
        result = self.check_read_data(field_data, 220, self.land_mask)
        self.assertMaskedArrayEqual(result, self.decomp_sea_data)

    def test_bad_lbpack(self):
        # Check basic land unpacking.
        field_data = self.sea_masked_data
        with self.assertRaises(ValueError):
            self.check_read_data(field_data, 320, self.land_mask)

    def check_read_data(self, field_data, lbpack, mask):
        # Calls pp._data_bytes_to_shaped_array with the necessary mocked
        # items, an lbpack instance, the correct data shape and mask instance.
        with mock.patch("numpy.frombuffer", return_value=field_data):
            data = pp._data_bytes_to_shaped_array(
                mock.Mock(),
                self.create_lbpack(lbpack),
                None,
                mask.shape,
                np.dtype(">f4"),
                -999,
                mask=mask,
            )
        return ma.masked_array(data, np.isnan(data), fill_value=-999)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Support for UM "fieldsfile-like" files.

At present, the only UM file types supported are true FieldsFiles and LBCs.
Other types of UM file may fail to load correctly (or at all).

"""

from iris.fileformats._ff import FF2PP
from iris.fileformats.pp import _load_cubes_variable_loader


def um_to_pp(filename, read_data=False, word_depth=None):
    """
    Extract individual PPFields from within a UM Fieldsfile-like file.

    Returns an iterator over the fields contained within the FieldsFile,
    returned as :class:`iris.fileformats.pp.PPField` instances.

    Args:

    * filename (string):
        Specify the name of the FieldsFile.

    Kwargs:

    * read_data (boolean):
        Specify whether to read the associated PPField data within
        the FieldsFile.  Default value is False.

    Returns:
        Iteration of :class:`iris.fileformats.pp.PPField`.

    For example::

        >>> for field in um.um_to_pp(filename):
        ...     print(field)

    """
    if word_depth is None:
        ff2pp = FF2PP(filename, read_data=read_data)
    else:
        ff2pp = FF2PP(filename, read_data=read_data, word_depth=word_depth)

    # Note: unlike the original wrapped case, we will return an actual
    # iterator, rather than an object that can provide an iterator.
    return iter(ff2pp)


def load_cubes(filenames, callback, constraints=None, _loader_kwargs=None):
    """
    Loads cubes from filenames of UM fieldsfile-like files.

    Args:

    * filenames - list of filenames to load

    Kwargs:

    * callback - a function which can be passed on to
        :func:`iris.io.run_callback`

    .. note::

        The resultant cubes may not be in the order that they are in the
        file (order is not preserved when there is a field with
        orography references).

    """
    return _load_cubes_variable_loader(
        filenames,
        callback,
        FF2PP,
        constraints=constraints,
        loading_function_kwargs=_loader_kwargs,
    )


def load_cubes_32bit_ieee(filenames, callback, constraints=None):
    """
    Loads cubes from filenames of 32bit ieee converted UM fieldsfile-like
    files.

    .. seealso::

        :func:`load_cubes` for keyword details

    """
    return load_cubes(
        filenames,
        callback,
        constraints=constraints,
        _loader_kwargs={"word_depth": 4},
    )



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
r"""
The purpose of this module is to provide utilities for the identification
of multi-dimensional structure in a flat sequence of homogeneous objects.
One application of this is to efficiently identify a higher dimensional
structure from a sorted sequence of PPField instances; for an example, given
a list of 12 PPFields, identification that there are 3 unique "time" values
and 4 unique "height" values where time and height are linearly independent
means that we could construct a resulting cube with a shape of
``(3, 4) + <shape of a single field>``.

An example using numpy arrays:

    >>> import numpy as np
    >>> orig_x, orig_y = np.arange(2), np.arange(3)
    >>> x, y = np.meshgrid(orig_x, orig_y)

    >>> # Remove the dimensional structure from the arrays.
    >>> x, y = x.flatten(), y.flatten()

    >>> print(x)
    [0 1 0 1 0 1]
    >>> print(y)
    [0 0 1 1 2 2]

    >>> arrays = {'x': x, 'y': y}
    >>> group = GroupStructure.from_component_arrays(arrays)
    >>> print(group)
    Group structure:
      Length: 6
      Element names: x, y
      Possible structures ("c" order):
        (y: 3; x: 2)

    >>> built_arrays = group.build_arrays((3, 2), arrays)
    >>> y_array, y_axes = built_arrays['y']
    >>> print(y_array, y_axes)
    [0 1 2] (0,)

"""

from collections import namedtuple

import numpy as np


class _UnstructuredArrayException(Exception):
    """
    Raised when an array has been incorrectly assumed to be
    structured in a specific way.

    """


class ArrayStructure(
    namedtuple("ArrayStructure", ["stride", "unique_ordered_values"])
):
    """
    Represents the identified structure of an array, where stride is the
    step between each unique value being seen in order in the flattened
    version of the array.

    Note: Stride is **not** in bytes, but is instead the number of objects in
    the original list of arrays, thus, stride is dtype independent.

    For column major (aka "F" order) arrays, stride will be one for those
    arrays which vary over the first dimension, and conversely will be one for
    C order arrays when varying over the last dimension.

    Constructing an ArrayStructure is most frequently done through the
    :meth:`ArrayStructure.from_array` class method, which takes a flattened
    array as its input.

    Stride examples:

    >>> ArrayStructure.from_array(np.array([1, 2])).stride
    1
    >>> ArrayStructure.from_array(np.array([1, 1, 2, 2])).stride
    2
    >>> ArrayStructure.from_array(np.array([1, 1, 1, 2, 2, 2])).stride
    3
    >>> ArrayStructure.from_array(np.array([1, 2, 1, 2])).stride
    1
    >>> ArrayStructure.from_array(np.array([1, 1, 2, 2, 1, 1, 2, 2])).stride
    2

    """

    def __new__(cls, stride, unique_ordered_values):
        self = super().__new__(cls, stride, unique_ordered_values)
        return self

    __slots__ = ()

    @property
    def size(self):
        """
        The ``size`` attribute is the number of the unique values in the
        original array. It is **not** the length of the original array.

        """
        return len(self.unique_ordered_values)

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        stride = getattr(other, "stride", None)
        arr = getattr(other, "unique_ordered_values", None)

        result = NotImplemented
        if stride is not None or arr is not None:
            result = stride == self.stride and np.all(
                self.unique_ordered_values == arr
            )
        return result

    def __ne__(self, other):
        return not (self == other)

    def construct_array(self, size):
        """
        The inverse operation of :func:`ArrayStructure.from_array`, returning
        a 1D array of the given length with the appropriate repetition
        pattern.

        """
        return np.tile(
            np.repeat(self.unique_ordered_values, self.stride),
            size // (self.size * self.stride),
        )

    def nd_array_and_dims(self, original_array, target_shape, order="c"):
        """
        Given a 1D array, and a target shape, construct an ndarray
        and associated dimensions.

        Raises an _UnstructuredArrayException if no optimised shape array can
        be returned, in which case, simply reshaping the original_array would
        be just as effective.

        For example:

        >>> orig = np.array([1, 2, 3, 1, 2, 3])
        >>> structure = ArrayStructure.from_array(orig)
        >>> array, dims = structure.nd_array_and_dims(orig, (2, 1, 3))
        >>> array
        array([1, 2, 3])
        >>> dims
        (2,)
        >>> # Filling the array with dimensions of length one should impact
        >>> # dims but not the array which is returned.
        >>> _, dims = structure.nd_array_and_dims(orig, (1, 2, 1, 3, 1))
        >>> dims
        (3,)

        """
        if original_array.shape[0] != np.prod(target_shape):
            raise ValueError(
                "Original array and target shape do not " "match up."
            )
        stride_product = 1

        result = None

        if self.size == 1:
            # There is no need to even consider the dimensionality - this
            # array structure only has one unique value, so it is a scalar
            # and has no associated dimension.
            result = (np.array(original_array[0]), ())

        for dim, length in sorted(
            enumerate(target_shape), reverse=order.lower() == "c"
        ):
            if result is not None:
                break

            # Does this array structure represent a flattened array of the
            # given shape? If so, reshape it back to the target shape,
            # then index out any dimensions which are constant.
            if self.stride == stride_product and length == self.size:
                vector = original_array.reshape(
                    target_shape + (-1,), order=order
                )
                # Reduce the dimensionality to a 1d array by indexing
                # everything but this dimension.
                vector = vector[
                    tuple(
                        0 if dim != i else slice(None)
                        for i in range(len(target_shape))
                    )
                ]
                # Remove any trailing dimension if it is trivial.
                if len(vector.shape) != 1 and vector.shape[-1] == 1:
                    vector = vector[..., 0]

                result = [vector, (dim,)]
                break

            stride_product *= length

        if result is not None:
            return result
        else:
            msg = (
                "Unable to construct an efficient nd_array for the target "
                "shape. Consider reshaping the array to the full shape "
                "instead."
            )
            raise _UnstructuredArrayException(msg)

    @classmethod
    def from_array(cls, arr):
        """
        Return the computed ArrayStructure for the given flat array
        (if a structure exists, otherwise return None).

        """
        # Note: This algorithm will only find distinct value columns/rows/axes
        # any dimension with repeat values will not have its structure
        # identified and will be considered irregular.

        if not isinstance(arr, np.ndarray):
            arr = np.array(arr)

        if arr.ndim != 1:
            raise ValueError("The given array must be 1D.")

        if arr.size == 0:
            return cls(1, arr)

        # unique is a *sorted* array of unique values.
        # unique_inds takes us from the sorted unique values back to inds in
        # the input array inds_back_to_orig gives us the indices of each value
        # in the array vs the index in the *sorted* unique array.
        _, unique_inds = np.unique(arr, return_index=True)

        # Return the unique values back into an ordered array.
        unique = arr[np.sort(unique_inds)]

        # what we actually want is inds_back_to_orig in the sort order of the
        # original array.
        new_inds = np.empty(arr.shape, dtype=unique_inds.dtype)

        for ind, unique_val in enumerate(unique):
            new_inds[arr == unique_val] = ind

        inds_back_to_orig = new_inds

        u_len = len(unique)
        n_fields = arr.size

        structure = None

        # If the length of the unique values is not a divisor of the
        # length of the original array, it is going to be an irregular
        # array, so we can avoid some processing.
        if (n_fields % u_len) != 0:
            # No structure.
            pass
        # Shortcut the simple case of all values being distinct.
        elif u_len == 1:
            structure = ArrayStructure(1, unique)
        else:
            # Working in index space, compute indices where values change.
            ind_diffs = np.diff(inds_back_to_orig)

            # Find the indices where a change takes place.
            ind_diffs_which_changed = np.nonzero(ind_diffs)[0]

            # Any index which changed by a different consecutive amount is a
            # stride. For example, an input array of [1,1,2,2,1,1,2,2]
            # results in ind_diffs looking like [0,1,0,-1,0,1,0] and
            # ind_diffs_which_changed being [1,3,5]. So now identifying the
            # "stride" (being the length of any sequence which has
            # consecutively equal values) is a matter of identifying the
            # difference between any two consecutive values from
            # ind_diffs_which_changed. If we don't have enough
            # ind_diffs_which_changed values to compute a difference, then
            # there is either one or two distinct values in the original
            # array, and the stride is therefore the total length / number
            # of unique values.
            try:
                stride = np.diff(ind_diffs_which_changed[:2])[0]
            except IndexError:
                stride = n_fields // u_len

            structure = cls(stride, unique)

            # Do one last sanity check - does the array we've just described
            # actually compute the correct array?
            constructed_array = structure.construct_array(arr.size)
            if not np.all(constructed_array == arr):
                structure = None

        return structure


class GroupStructure:
    """
    The GroupStructure class represents a collection of array structures along
    with additional information such as the length of the arrays and the array
    order in which they are found (row-major or column-major).

    """

    def __init__(self, length, component_structure, array_order="c"):
        """
        group_component_to_array - a dictionary. See also TODO
        """
        #: The size common to all of the original arrays and used to determine
        #: possible shape configurations.
        self.length = length

        #: A dictionary mapping component name to ArrayStructure instance
        #: (or None if no such structure exists for that component).
        self._cmpt_structure = component_structure

        array_order = array_order.lower()
        if array_order not in ["c", "f"]:
            raise ValueError("Invalid array order {!r}".format(array_order))
        self._array_order = array_order

    @classmethod
    def from_component_arrays(cls, component_arrays, array_order="c"):
        """
        Given a dictionary of component name to flattened numpy array,
        return an :class:`GroupStructure` instance which is representative
        of the underlying array structures.

        Args:

         * component_arrays - A dictionary mapping component name to the
                              full sized 1d (flattened) numpy array.

        """
        cmpt_structure = {
            name: ArrayStructure.from_array(array)
            for name, array in component_arrays.items()
        }

        sizes = np.array([array.size for array in component_arrays.values()])
        if not np.all(sizes == sizes[0]):
            raise ValueError("All array elements must have the same size.")

        return cls(sizes[0], cmpt_structure, array_order=array_order)

    def _potentially_flattened_components(self):
        """
        Return a generator of the components which could form non-trivial
        (i.e. ``length > 1``) array dimensions.

        """
        for name, structure in self._cmpt_structure.items():
            if structure is not None and structure.size > 1:
                yield (name, structure)

    @property
    def is_row_major(self):
        return self._array_order == "c"

    def possible_structures(self):
        """
        Return a tuple containing the possible structures that this group
        could have.

        A structure in this case is an iterable of
        ``(name, ArrayStructure)`` pairs, one per dimension, of a possible
        array. The shape of the resulting array would then be
        ``tuple(array_struct.size for (name, array_struct) in pair)`` for any
        of the returned structures.

        The algorithm does not deal with incomplete structures, such that
        all components critical for the identification of a shape are
        necessary.

        """
        vector_structures = sorted(self._potentially_flattened_components())

        def filter_strides_of_length(length):
            return [
                (name, struct)
                for (name, struct) in vector_structures
                if struct.stride == length
            ]

        # Keep track of our structures so far. This will be a list of lists
        # containing the actual structure pairs.
        possible = []

        # Get hold of all array structures with a stride of 1. These are the
        # only possible array structures for the first dimension (i.e. on the
        # left-most dimension for column-major ordering). Start a list of
        # structures, one for each possible left hand side dimension
        # component.
        for structure in filter_strides_of_length(1):
            possible.append([structure])

        # Keep track of all structures which are valid, these are ultimately
        # what will be returned from this function.
        allowed_structures = []

        # We will make use of the mutability of the possible list, removing
        # each list representing a potential structure. With the potential
        # just removed, we will find any array structures which could be the
        # next dimension in the potential, adding one new possible structure
        # per array structure found. If at any point, a possible's stride
        # product is the same as the length of this group, we've got an
        # allowed structure.
        while possible:
            for potential in possible[:]:
                possible.remove(potential)
                # If we are to build another dimension on top of this possible
                # structure, we need to compute the stride that would be
                # needed for that dimension.
                next_stride = np.product(
                    [struct.size for (_, struct) in potential]
                )

                # If we've found a structure whose product is the length of
                # the fields of this Group, we've got a valid potential.
                if next_stride == self.length:
                    allowed_structures.append(potential)

                # So let's get all of the potential nd-arrays which would be
                # viable dimensions on this potential structure.
                next_dim_structs = filter_strides_of_length(next_stride)

                # Any we find get added to this potential and put back in the
                # possibles.
                for struct in next_dim_structs:
                    if struct in potential:
                        continue
                    new_potential = potential[:]
                    # Add the structure to the potential on the right hand
                    # side of the structure (column-major).
                    new_potential.append(struct)
                    possible.append(new_potential)

        # We've been working in column-major order, so let's reverse the
        # dimensionality if we are in row-major.
        if self.is_row_major:
            for potential in allowed_structures:
                potential.reverse()

        return tuple(allowed_structures)

    def __str__(self):
        result = [
            "Group structure:",
            "  Length: {}".format(self.length),
            "  Element names: {}".format(
                ", ".join(sorted(self._cmpt_structure.keys()))
            ),
            '  Possible structures ("{}" order):'.format(self._array_order),
        ]

        for structure in self.possible_structures():
            sizes = (
                "{}: {}".format(name, arr_struct.size)
                for name, arr_struct in structure
            )
            result.append("    ({})".format("; ".join(sizes)))

        return "\n".join(result)

    def build_arrays(self, shape, elements_arrays):
        """
        Given the target shape, and a dictionary mapping name to 1D array of
        :attr:`.length`, return a dictionary mapping element name to
        ``(ndarray, dims)``.

        Note: Actually the arrays may be more than one dimension, and the
        trailing dimensions will be preserved. This is useful for items such
        as datetimes, where an efficiency exists to avoid the construction
        of datetime objects until the last moment.

        """
        elem_to_nd_and_dims = {}

        if sorted(elements_arrays.keys()) != sorted(self._cmpt_structure):
            raise ValueError(
                "The GroupStructure elements were not the same "
                "as those provided in the element_arrays."
            )

        for name, array in elements_arrays.items():
            struct = self._cmpt_structure[name]
            nd_array_and_dims = None
            if struct is not None:
                try:
                    nd_array_and_dims = struct.nd_array_and_dims(
                        array, shape, order=self._array_order
                    )
                except _UnstructuredArrayException:
                    pass

            if nd_array_and_dims is None:
                reshape_shape = shape
                if array.ndim > 1:
                    reshape_shape = reshape_shape + (-1,)
                nd_array_and_dims = [
                    array.reshape(reshape_shape, order=self._array_order),
                    tuple(range(len(shape))),
                ]
            elem_to_nd_and_dims[name] = nd_array_and_dims
        return elem_to_nd_and_dims
